import unittest
from collections import OrderedDict

from hamlpy.compiler import Compiler
from hamlpy.parser.core import Stream
from hamlpy.parser.elements import Element, read_element, read_tag


class ElementTest(unittest.TestCase):
    def test_read_tag(self):
        stream = Stream("angular:ng-repeat(")
        self.assertEqual(read_tag(stream), "angular:ng-repeat")
        self.assertEqual(stream.text[stream.ptr :], "(")

    def test_read_element(self):
        stream = Stream('%angular:ng-repeat.my-class.other:class#my-id(class="test")></=  Hello  \n.next-element')
        element = read_element(stream, Compiler())
        self.assertEqual(element.tag, "angular:ng-repeat")
        self.assertEqual(element.id, "my-id")
        self.assertEqual(element.classes, ["test", "my-class", "other:class"])
        self.assertEqual(dict(element.attributes), {"class": "test"})
        self.assertEqual(element.nuke_outer_whitespace, True)
        self.assertEqual(element.nuke_inner_whitespace, True)
        self.assertEqual(element.self_close, True)
        self.assertEqual(element.django_variable, True)
        self.assertEqual(element.inline_content, "Hello")
        self.assertEqual(stream.text[stream.ptr :], ".next-element")

        stream = Stream("%input{required}  ")
        element = read_element(stream, Compiler())
        self.assertEqual(element.tag, "input")
        self.assertEqual(element.id, None)
        self.assertEqual(element.classes, [])
        self.assertEqual(dict(element.attributes), {"required": True})
        self.assertEqual(element.nuke_outer_whitespace, False)
        self.assertEqual(element.nuke_inner_whitespace, False)
        self.assertEqual(element.self_close, True)  # input is implicitly self-closing
        self.assertEqual(element.django_variable, False)
        self.assertEqual(element.inline_content, "")
        self.assertEqual(stream.text[stream.ptr :], "")

    def test_escape_attribute_quotes(self):
        s1 = Element._escape_attribute_quotes("""{% url 'blah' %}""", attr_wrapper="'")
        assert s1 == """{% url 'blah' %}"""

        s2 = Element._escape_attribute_quotes(
            """'foo'="bar" {% url 'blah' "blah" %} blah's blah''s""", attr_wrapper="'"
        )
        assert s2 == """&#39;foo&#39;="bar" {% url 'blah' "blah" %} blah&#39;s blah&#39;&#39;s"""

        s3 = Element._escape_attribute_quotes("""'foo'="bar" {% url 'blah' "blah" %}""", attr_wrapper='"')
        assert s3 == """'foo'=&quot;bar&quot; {% url 'blah' "blah" %}"""

    def test_parses_tag(self):
        element = self._read_element("%span.class")
        assert element.tag == "span"

        # can have namespace and hypens
        element = self._read_element("%ng-tags:ng-repeat.class")
        assert element.tag == "ng-tags:ng-repeat"

        # defaults to div
        element = self._read_element(".class#id")
        assert element.tag == "div"

    def test_parses_id(self):
        element = self._read_element("%div#someId.someClass")
        assert element.id == "someId"

        element = self._read_element("#someId.someClass")
        assert element.id == "someId"

        element = self._read_element("%div.someClass")
        assert element.id is None

    def test_parses_classes(self):
        element = self._read_element("%div#someId.someClass")
        assert element.classes == ["someClass"]

        element = self._read_element("%div#someId.someClass.anotherClass")
        assert element.classes == ["someClass", "anotherClass"]

        # classes before id
        element = self._read_element("%div.someClass.anotherClass#someId")
        assert element.classes == ["someClass", "anotherClass"]

        # classes can contain hypens, underscores and colons
        element = self._read_element("%div.-some-class-._another_class_.also:class")
        assert element.classes == ["-some-class-", "_another_class_", "also:class"]

        # no class
        element = self._read_element("%div#someId")
        assert element.classes == []

    def test_attribute_dictionary_properly_parses(self):
        element = self._read_element("%html{'xmlns':'http://www.w3.org/1999/xhtml', 'xml:lang':'en', 'lang':'en'}")

        assert element.attributes == OrderedDict(
            [("xmlns", "http://www.w3.org/1999/xhtml"), ("xml:lang", "en"), ("lang", "en")]
        )

    def test_attribute_merges_classes_properly(self):
        element = self._read_element("%div.someClass.anotherClass{'class':'hello'}")

        assert element.classes == ["hello", "someClass", "anotherClass"]

    def test_attribute_merges_ids_properly(self):
        element = self._read_element("%div#someId{'id':'hello'}")
        assert element.id == "someId_hello"

    def test_can_use_arrays_for_id_in_attributes(self):
        element = self._read_element("%div#someId{'id':['more', 'andMore']}")
        assert element.id == "someId_more_andMore"

    def test_self_closes_a_self_closing_tag(self):
        element = self._read_element(r"%br")
        assert element.self_close

    def test_does_not_close_a_non_self_closing_tag(self):
        element = self._read_element("%div")
        assert element.self_close is False

    def test_can_close_a_non_self_closing_tag(self):
        element = self._read_element("%div/")
        assert element.self_close

    def test_properly_detects_django_tag(self):
        element = self._read_element("%div= $someVariable")
        assert element.django_variable

    def test_knows_when_its_not_django_tag(self):
        element = self._read_element("%div Some Text")
        assert element.django_variable is False

    def test_grabs_inline_tag_content(self):
        element = self._read_element("%div Some Text")
        assert element.inline_content == "Some Text"

        element = self._read_element("%div {% trans 'hello' %}")
        assert element.inline_content == "{% trans 'hello' %}"

    def test_multiline_attributes(self):
        element = self._read_element(
            """%link{'rel': 'stylesheet', 'type': 'text/css',
            'href': '/long/url/to/stylesheet/resource.css'}"""
        )

        assert element.attributes == OrderedDict(
            [("rel", "stylesheet"), ("type", "text/css"), ("href", "/long/url/to/stylesheet/resource.css")]
        )

    @staticmethod
    def _read_element(text):
        return read_element(Stream(text), Compiler())



import unittest
from collections import OrderedDict

from hamlpy.compiler import Compiler
from hamlpy.parser.attributes import read_attribute_dict
from hamlpy.parser.core import ParseException, Stream


class AttributeDictParserTest(unittest.TestCase):
    @staticmethod
    def _parse(text):
        return read_attribute_dict(Stream(text), Compiler())

    def test_read_ruby_style_attribute_dict(self):
        # empty dict
        stream = Stream("{}><")
        assert dict(read_attribute_dict(stream, Compiler())) == {}
        assert stream.text[stream.ptr :] == "><"

        # string values
        assert dict(self._parse("{'class': 'test'} =Test")) == {"class": "test"}
        assert dict(self._parse("{'class': 'test', 'id': 'something'}")) == {"class": "test", "id": "something"}

        # integer values
        assert dict(self._parse("{'data-number': 0}")) == {"data-number": "0"}
        assert dict(self._parse("{'data-number': 12345}")) == {"data-number": "12345"}

        # float values
        assert dict(self._parse("{'data-number': 123.456}")) == {"data-number": "123.456"}
        assert dict(self._parse("{'data-number': 0.001}")) == {"data-number": "0.001"}

        # None value
        assert dict(self._parse("{'controls': None}")) == {"controls": None}

        # boolean attributes
        assert dict(self._parse("{disabled, class:'test', data-number : 123,\n foo:\"bar\"}")) == {
            "disabled": True,
            "class": "test",
            "data-number": "123",
            "foo": "bar",
        }

        assert dict(self._parse("{class:'test', data-number : 123,\n foo:\"bar\",  \t   disabled}")) == {
            "disabled": True,
            "class": "test",
            "data-number": "123",
            "foo": "bar",
        }

        # attribute name has colon
        assert dict(self._parse("{'xml:lang': 'en'}")) == {"xml:lang": "en"}

        # attribute value has colon or commas
        assert dict(self._parse("{'lang': 'en:g'}")) == {"lang": "en:g"}
        assert dict(
            self._parse(
                '{name:"viewport", content:"width:device-width, initial-scale:1, minimum-scale:1, maximum-scale:1"}'
            )
        ) == {"name": "viewport", "content": "width:device-width, initial-scale:1, minimum-scale:1, maximum-scale:1"}

        # double quotes
        assert dict(self._parse('{"class": "test", "id": "something"}')) == {"class": "test", "id": "something"}

        # no quotes for key
        assert dict(self._parse("{class: 'test', id: 'something'}")) == {"class": "test", "id": "something"}

        # whitespace is ignored
        assert dict(self._parse("{   class  \t :        'test',     data-number:    123  }")) == {
            "class": "test",
            "data-number": "123",
        }

        # trailing commas are fine
        assert dict(self._parse("{class: 'test', data-number: 123,}")) == {"class": "test", "data-number": "123"}

        # attributes split onto multiple lines
        assert dict(self._parse("{class: 'test',\n     data-number: 123}")) == {"class": "test", "data-number": "123"}

        # old style Ruby
        assert dict(self._parse("{:class => 'test', :data-number=>123}")) == {"class": "test", "data-number": "123"}

        # list attribute values
        assert dict(self._parse("{'class': [ 'a', 'b', 'c' ], data-list:[1, 2, 3]}")) == {
            "class": ["a", "b", "c"],
            "data-list": ["1", "2", "3"],
        }

        # tuple attribute values
        assert dict(self._parse("{:class=>( 'a', 'b', 'c' ), :data-list => (1, 2, 3)}")) == {
            "class": ["a", "b", "c"],
            "data-list": ["1", "2", "3"],
        }

        # attribute order is maintained
        assert self._parse("{'class': 'test', 'id': 'something', foo: 'bar'}") == OrderedDict(
            [("class", "test"), ("id", "something"), ("foo", "bar")]
        )

        # attribute values can be multi-line Haml
        haml = """{
                'class':
                    - if forloop.first
                        link-first
\x20
                    - else
                        - if forloop.last
                            link-last
                'href':
                    - url 'some_view'
        }"""
        assert dict(self._parse(haml)) == {
            "class": "{% if forloop.first %} link-first {% else %} {% if forloop.last %} link-last {% endif %} {% endif %}",  # noqa
            "href": "{% url 'some_view' %}",
        }

        # non-ascii attribute values
        assert dict(self._parse("{class: 'test\u1234'}")) == {"class": "test\u1234"}

    def test_read_html_style_attribute_dict(self):
        # html style dicts
        assert dict(self._parse("()><")) == {}
        assert dict(self._parse("(   )")) == {}

        # string values
        assert dict(self._parse("(class='test') =Test")) == {"class": "test"}
        assert dict(self._parse("(class='test' id='something')")) == {"class": "test", "id": "something"}

        # integer values
        assert dict(self._parse("(data-number=0)")) == {"data-number": "0"}
        assert dict(self._parse("(data-number=12345)")) == {"data-number": "12345"}

        # float values
        assert dict(self._parse("(data-number=123.456)")) == {"data-number": "123.456"}
        assert dict(self._parse("(data-number=0.001)")) == {"data-number": "0.001"}

        # None value
        assert dict(self._parse("(controls=None)")) == {"controls": None}

        # boolean attributes
        assert dict(self._parse("(disabled class='test' data-number = 123\n foo=\"bar\")")) == {
            "disabled": True,
            "class": "test",
            "data-number": "123",
            "foo": "bar",
        }

        assert dict(self._parse("(class='test' data-number = 123\n foo=\"bar\"  \t   disabled)")) == {
            "disabled": True,
            "class": "test",
            "data-number": "123",
            "foo": "bar",
        }

        # attribute name has colon
        assert dict(self._parse('(xml:lang="en")')) == {"xml:lang": "en"}

        # attribute names with characters found in JS frameworks
        assert dict(self._parse('([foo]="a" ?foo$="b")')) == {"[foo]": "a", "?foo$": "b"}

        # double quotes
        assert dict(self._parse('(class="test" id="something")')) == {"class": "test", "id": "something"}

        # list attribute values
        assert dict(self._parse("(class=[ 'a', 'b', 'c' ] data-list=[1, 2, 3])")) == {
            "class": ["a", "b", "c"],
            "data-list": ["1", "2", "3"],
        }

        # variable attribute values
        assert dict(self._parse("(foo=bar)")) == {"foo": "{{ bar }}"}

        # attribute values can be multi-line Haml
        haml = """(
                class=
                    - if forloop.first
                        link-first
\x20
                    - else
                        - if forloop.last
                            link-last
                href=
                    - url 'some_view'
        )"""
        assert dict(self._parse(haml)) == {
            "class": "{% if forloop.first %} link-first {% else %} {% if forloop.last %} link-last {% endif %} {% endif %}",  # noqa
            "href": "{% url 'some_view' %}",
        }

    def test_empty_attribute_name_raises_error(self):
        # empty quoted string in Ruby new style
        with self.assertRaisesRegex(ParseException, r'Attribute name can\'t be an empty string. @ "{\'\':" <-'):
            self._parse("{'': 'test'}")

        # empty old style Ruby attribute
        with self.assertRaisesRegex(ParseException, r'Unexpected " ". @ "{: " <-'):
            self._parse("{: 'test'}")

        # missing (HTML style)
        with self.assertRaisesRegex(ParseException, r'Unexpected "=". @ "\(=" <-'):
            self._parse("(='test')")
        with self.assertRaisesRegex(ParseException, r'Unexpected "=". @ "\(foo=\'bar\' =" <-'):
            self._parse("(foo='bar' ='test')")

    def test_empty_attribute_value_raises_error(self):
        with self.assertRaisesRegex(ParseException, r'Unexpected "}". @ "{:class=>}" <-'):
            self._parse("{:class=>}")
        with self.assertRaisesRegex(ParseException, r'Unexpected "}". @ "{class:}" <-'):
            self._parse("{class:}")
        with self.assertRaisesRegex(ParseException, r'Unexpected "\)". @ "\(class=\)" <-'):
            self._parse("(class=)")

    def test_unterminated_string_raises_error(self):
        # on attribute key
        with self.assertRaisesRegex(ParseException, r'Unterminated string \(expected \'\). @ "{\'test: 123}" <-'):
            self._parse("{'test: 123}")

        # on attribute value
        with self.assertRaisesRegex(ParseException, r'Unterminated string \(expected "\). @ "{\'test\': "123}" <-'):
            self._parse("{'test': \"123}")

    def test_duplicate_attributes_raise_error(self):
        with self.assertRaisesRegex(
            ParseException, r'Duplicate attribute: "class". @ "{class: \'test\', class: \'bar\'}" <-'
        ):  # noqa
            self._parse("{class: 'test', class: 'bar'}")

        with self.assertRaisesRegex(
            ParseException, r'Duplicate attribute: "class". @ "\(class=\'test\' class=\'bar\'\)" <-'
        ):  # noqa
            self._parse("(class='test' class='bar')")

    def test_mixing_ruby_and_html_syntax_raises_errors(self):
        # omit comma in Ruby style dict
        with self.assertRaisesRegex(ParseException, r'Expected ",". @ "{class: \'test\' f" <-'):
            self._parse("{class: 'test' foo: 'bar'}")

        # use = in Ruby style dict
        with self.assertRaisesRegex(ParseException, r'Expected ":". @ "{class=" <-'):
            self._parse("{class='test'}")
        with self.assertRaisesRegex(ParseException, r'Expected "=>". @ "{:class=" <-'):
            self._parse("{:class='test'}")

        # use colon as assignment for old style Ruby attribute
        with self.assertRaisesRegex(ParseException, r'Expected "=>". @ "{:class:" <-'):
            self._parse("{:class:'test'}")

        # use comma in HTML style dict
        with self.assertRaisesRegex(ParseException, r'Unexpected ",". @ "\(class=\'test\'," <-'):
            self._parse("(class='test', foo = 'bar')")

        # use : for assignment in HTML style dict (will treat as part of attribute name)
        with self.assertRaisesRegex(ParseException, r'Unexpected "\'". @ "\(class:\'" <-'):
            self._parse("(class:'test')")

        # use attribute quotes in HTML style dict
        with self.assertRaisesRegex(ParseException, r'Unexpected "\'". @ "\(\'" <-'):
            self._parse("('class'='test')")

        # use => in HTML style dict
        with self.assertRaisesRegex(ParseException, r'Unexpected ">". @ "\(class=>" <-'):
            self._parse("(class=>'test')")

        # use tuple syntax in HTML style dict
        with self.assertRaisesRegex(ParseException, r'Unexpected "\(". @ "\(class=\(" <-'):
            self._parse("(class=(1, 2))")

    def test_unexpected_eof(self):
        with self.assertRaisesRegex(ParseException, r'Unexpected end of input. @ "{:class=>" <-'):
            self._parse("{:class=>")
        with self.assertRaisesRegex(ParseException, r'Unexpected end of input. @ "{class:" <-'):
            self._parse("{class:")
        with self.assertRaisesRegex(ParseException, r'Unexpected end of input. @ "\(class=" <-'):
            self._parse("(class=")



"""
This module decorates the django templatize function to parse haml templates before the translation utility extracts
tags from it.
"""

import os

from django.utils import translation

from hamlpy import HAML_EXTENSIONS
from hamlpy.compiler import Compiler


def patch_templatize(func):
    def templatize(src, origin=None, charset="utf-8"):
        # if the template has no origin then don't attempt to convert it because we don't know if it's Haml
        if origin:
            extension = os.path.splitext(origin)[1][1:].lower()

            if extension in HAML_EXTENSIONS:
                compiler = Compiler()
                src = compiler.process(src)

        return func(src, origin=origin)

    return templatize


translation.templatize = patch_templatize(translation.templatize)



"""
Based on original haml-watcher.py by Christian Stefanescu (st.chris@gmail.com)

Watches a folder for files with the given extensions and calls the HamlPy compiler if the modified time has changed
since the last check.
"""

import argparse
import codecs
import os
import os.path
import sys
import time
from time import strftime

from hamlpy import HAML_EXTENSIONS
from hamlpy.compiler import Compiler


class Options(object):
    CHECK_INTERVAL = 3  # in seconds
    DEBUG = False  # print file paths when a file is compiled
    VERBOSE = False
    OUTPUT_EXT = ".html"


# dict of compiled files, fullpath => timestamp
compiled = dict()


class StoreNameValueTagPair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        tags = getattr(namespace, "tags", {})
        for item in values:
            n, v = item.split(":")
            tags[n] = v

        setattr(namespace, "tags", tags)


arg_parser = argparse.ArgumentParser()
arg_parser.add_argument("-v", "--verbose", help="Display verbose output", action="store_true")
arg_parser.add_argument(
    "-i", "--input-extension", metavar="EXT", help="The file extensions to look for.", type=str, nargs="+"
)
arg_parser.add_argument(
    "-ext",
    "--extension",
    metavar="EXT",
    default=Options.OUTPUT_EXT,
    help="The output file extension. Default is .html",
    type=str,
)
arg_parser.add_argument(
    "-r",
    "--refresh",
    metavar="S",
    default=Options.CHECK_INTERVAL,
    type=int,
    help="Refresh interval for files. Default is %d seconds. Ignored if the --once flag is set."
    % Options.CHECK_INTERVAL,
)
arg_parser.add_argument("input_dir", help="Folder to watch", type=str)
arg_parser.add_argument("output_dir", help="Destination folder", type=str, nargs="?")
arg_parser.add_argument(
    "--tag", type=str, nargs=1, action=StoreNameValueTagPair, help="Add self closing tag. eg. --tag macro:endmacro"
)
arg_parser.add_argument(
    "--attr-wrapper",
    dest="attr_wrapper",
    type=str,
    choices=('"', "'"),
    default="'",
    action="store",
    help="The character that should wrap element attributes. This defaults to ' (an apostrophe).",
)
arg_parser.add_argument(
    "--django-inline",
    dest="django_inline",
    action="store_true",
    help="Whether to support ={...} syntax for inline variables in addition to #{...}",
)
arg_parser.add_argument(
    "--jinja", default=False, action="store_true", help="Makes the necessary changes to be used with Jinja2."
)
arg_parser.add_argument(
    "--once",
    default=False,
    action="store_true",
    help="Runs the compiler once and exits on completion. "
    "Returns a non-zero exit code if there were any compile errors.",
)


def watch_folder():
    """Main entry point. Expects one or two arguments (the watch folder + optional destination folder)."""
    args = arg_parser.parse_args(sys.argv[1:])
    compiler_args = {}

    input_folder = os.path.realpath(args.input_dir)
    if not args.output_dir:
        output_folder = input_folder
    else:
        output_folder = os.path.realpath(args.output_dir)

    if args.verbose:
        Options.VERBOSE = True
        print("Watching {} at refresh interval {} seconds".format(input_folder, args.refresh))

    if args.extension:
        Options.OUTPUT_EXT = args.extension

    if getattr(args, "tags", False):
        compiler_args["custom_self_closing_tags"] = args.tags

    if args.input_extension:
        input_extensions = [(e[1:] if e.startswith(".") else e) for e in args.input_extension]  # strip . chars
    else:
        input_extensions = HAML_EXTENSIONS

    if args.attr_wrapper:
        compiler_args["attr_wrapper"] = args.attr_wrapper

    if args.django_inline:
        compiler_args["django_inline_style"] = args.django_inline

    if args.jinja:
        compiler_args["tag_config"] = "jinja2"

    # compile once, then exist
    if args.once:
        (total_files, num_failed) = _watch_folder(input_folder, input_extensions, output_folder, compiler_args)
        print("Compiled %d of %d files." % (total_files - num_failed, total_files))
        if num_failed == 0:
            print("All files compiled successfully.")
        else:
            print("Some files have errors.")
        sys.exit(num_failed)

    while True:
        try:
            _watch_folder(input_folder, input_extensions, output_folder, compiler_args)
            time.sleep(args.refresh)
        except KeyboardInterrupt:
            # allow graceful exit (no stacktrace output)
            sys.exit(0)


def _watch_folder(folder, extensions, destination, compiler_args):
    """
    Compares "modified" timestamps against the "compiled" dict, calls compiler if necessary. Returns a tuple of the
    number of files hit and the number of failed compiles
    """
    total_files = 0
    num_failed = 0

    if Options.VERBOSE:
        print("_watch_folder(%s, %s, %s)" % (folder, repr(extensions), destination))

    for dirpath, dirnames, filenames in os.walk(folder):
        for filename in filenames:
            # ignore filenames starting with ".#" for Emacs compatibility
            if filename.startswith(".#"):
                continue

            if not _has_extension(filename, extensions):
                continue

            fullpath = os.path.join(dirpath, filename)
            subfolder = os.path.relpath(dirpath, folder)
            mtime = os.stat(fullpath).st_mtime

            # create subfolders in target directory if they don't exist
            compiled_folder = os.path.join(destination, subfolder)
            if not os.path.exists(compiled_folder):
                os.makedirs(compiled_folder)

            compiled_path = _compiled_path(compiled_folder, filename)
            if fullpath not in compiled or compiled[fullpath] < mtime or not os.path.isfile(compiled_path):
                compiled[fullpath] = mtime
                total_files += 1
                if not compile_file(fullpath, compiled_path, compiler_args):
                    num_failed += 1

    return total_files, num_failed


def _has_extension(filename, extensions):
    """
    Checks if the given filename has one of the given extensions
    """
    for ext in extensions:
        if filename.endswith("." + ext):
            return True
    return False


def _compiled_path(destination, filename):
    return os.path.join(destination, filename[: filename.rfind(".")] + Options.OUTPUT_EXT)


def compile_file(fullpath, outfile_name, compiler_args):
    """
    Calls HamlPy compiler. Returns True if the file was compiled and written successfully.
    """
    if Options.VERBOSE:
        print("%s %s -> %s" % (strftime("%H:%M:%S"), fullpath, outfile_name))
    try:
        if Options.DEBUG:  # pragma: no cover
            print("Compiling %s -> %s" % (fullpath, outfile_name))
        haml = codecs.open(fullpath, "r", encoding="utf-8").read()
        compiler = Compiler(compiler_args)
        output = compiler.process(haml)
        outfile = codecs.open(outfile_name, "w", encoding="utf-8")
        outfile.write(output)

        return True
    except Exception as e:
        # import traceback
        print("Failed to compile %s -> %s\nReason:\n%s" % (fullpath, outfile_name, e))
        # print traceback.print_exc()

    return False


if __name__ == "__main__":  # pragma: no cover
    watch_folder()



#!/usr/bin/env python

import warnings

import regex

from hamlpy.parser.core import Stream
from hamlpy.parser.nodes import Node, read_node


class Options(object):
    HTML4 = "html4"
    HTML5 = "html5"
    XHTML = "xhtml"

    def __init__(self, **kwargs):
        # standard Haml options
        self.attr_wrapper = "'"  # how to render attribute values, e.g. foo='bar'
        self.format = self.HTML5  # HTML4, HTML5 or XHTML
        self.escape_attrs = False  # escape HTML-sensitive characters in attribute values
        self.cdata = False  # wrap CSS, Javascript etc content in CDATA section

        # implementation specific options
        self.django_inline_style = False  # support both #{...} and ={...}
        self.tag_config = "django"  # Django vs Jinja2 tags
        self.custom_self_closing_tags = {}  # additional self-closing tags
        self.endblock_names = False  # include block name on endblock closing tags
        self.smart_quotes = False
        self.debug_tree = False

        for k, v in kwargs.items():
            setattr(self, k, v)

        if self.django_inline_style:  # pragma: no cover
            warnings.warn("Support for ={..} style variables is deprecated", DeprecationWarning)

        self.cdata = self._cdata

    @property
    def html4(self):
        return self.format == self.HTML4

    @property
    def html5(self):
        return self.format == self.HTML5

    @property
    def html(self):
        return self.html4 or self.html5

    @property
    def xhtml(self):
        return self.format == self.XHTML

    @property
    def _cdata(self):
        return self.xhtml or self.cdata


class Compiler:
    TAG_CONFIGS = {
        "django": {
            "self_closing": {
                "for": "endfor",
                "if": "endif",
                "ifchanged": "endifchanged",
                "ifequal": "endifequal",
                "ifnotequal": "endifnotequal",
                "block": "endblock",
                "filter": "endfilter",
                "autoescape": "endautoescape",
                "with": "endwith",
                "blocktrans": "endblocktrans",
                "spaceless": "endspaceless",
                "comment": "endcomment",
                "cache": "endcache",
                "localize": "endlocalize",
                "call": "endcall",
                "macro": "endmacro",
                "compress": "endcompress",
            },
            "may_contain": {
                "blocktrans": ["plural"],
                "if": ["else", "elif"],
                "ifchanged": ["else"],
                "ifequal": ["else"],
                "ifnotequal": ["else"],
                "for": ["empty"],
            },
        },
        "jinja2": {
            "self_closing": {
                "for": "endfor",
                "if": "endif",
                "block": "endblock",
                "filter": "endfilter",
                "with": "endwith",
                "call": "endcall",
                "macro": "endmacro",
                "raw": "endraw",
            },
            "may_contain": {"if": ["else", "elif"], "for": ["empty", "else"]},
        },
    }

    def __init__(self, options=None):
        self.options = Options(**(options or {}))

        tag_config = self.TAG_CONFIGS[self.options.tag_config]
        self.self_closing_tags = tag_config["self_closing"]
        self.tags_may_contain = tag_config["may_contain"]

        self.self_closing_tags.update(self.options.custom_self_closing_tags)

        self.inline_variable_regexes = self._create_inline_variable_regexes()

    def process(self, haml):
        """
        Converts the given string of Haml to a regular Django HTML
        """
        stream = Stream(haml)

        root = Node.create_root(self)
        node = None

        while True:
            node = read_node(stream, prev=node, compiler=self)
            if not node:
                break

            root.add_node(node)

        if self.options.debug_tree:  # pragma: no cover
            return root.debug_tree()
        else:
            return root.render()

    def _create_inline_variable_regexes(self):
        """
        Generates regular expressions for inline variables and escaped inline variables, based on compiler options
        """
        prefixes = ["=", "#"] if self.options.django_inline_style else ["#"]
        prefixes = "".join(prefixes)
        return (
            regex.compile(r"(?<!\\)([" + prefixes + r"]\{\s*(.+?)\s*\})"),
            regex.compile(r"\\([" + prefixes + r"]\{\s*(.+?)\s*\})"),
        )



import unittest

from hamlpy.views.generic import CreateView, DetailView, UpdateView


class DummyCreateView(CreateView):
    template_name = "create.html"


class DummyDetailView(DetailView):
    template_name = "detail.htm"


class DummyUpdateView(UpdateView):
    template_name = "update.xml"


class DjangoViewsTest(unittest.TestCase):
    def test_get_template_names(self):
        assert DummyCreateView().get_template_names() == ["create.haml", "create.hamlpy", "create.html"]
        assert DummyDetailView().get_template_names() == ["detail.haml", "detail.hamlpy", "detail.htm"]
        assert DummyUpdateView().get_template_names() == ["update.haml", "update.hamlpy", "update.xml"]



import unittest

from jinja2.exceptions import TemplateSyntaxError

from hamlpy.jinja import HamlPyExtension


class JinjaTest(unittest.TestCase):
    def test_preprocess(self):
        extension = HamlPyExtension(None)

        assert extension.preprocess("%span Hello", "test.haml") == "<span>Hello</span>\n"
        assert extension.preprocess("%span Hello", "test.hamlpy") == "<span>Hello</span>\n"
        assert extension.preprocess("%span Hello", "../templates/test.hamlpy") == "<span>Hello</span>\n"

        # non-Haml extension should be ignored
        assert extension.preprocess("%span Hello", "test.txt") == "%span Hello"

        # exceptions converted to Jinja2 exceptions
        self.assertRaisesRegex(
            TemplateSyntaxError,
            r'Unterminated string \(expected "\). @ "%span{"}"',
            extension.preprocess,
            '%span{"}',
            "test.haml",
        )



from enum import Enum

from .core import Stream

DJANGO_TAG_OPEN = "{%"
DJANGO_TAG_CLOSE = "%}"
DJANGO_EXP_OPEN = "{{"
DJANGO_EXP_CLOSE = "}}"
HTML_CHARS = {"&": "&amp;", "<": "&lt;", ">": "&gt;", '"': "&quot;", "'": "&#39;"}


class EscapeState(Enum):
    normal = 0
    in_tag = 1
    in_exp = 2


def html_escape(text):
    """
    Escapes HTML entities, matching substitutions used by the Ruby Haml library. Entities that occur inside Django tags
    or expressions are not escaped.
    """
    new_text = []
    state = EscapeState.normal
    stream = Stream(text)

    while stream.ptr < stream.length:
        ch = stream.text[stream.ptr]
        ch2 = stream.text[stream.ptr : stream.ptr + 2]

        if ch2 == DJANGO_TAG_OPEN or ch2 == DJANGO_EXP_OPEN:
            state = EscapeState.in_tag if ch2 == DJANGO_TAG_OPEN else EscapeState.in_exp
            stream.ptr += 2
            new_text.append(ch2)
        elif (state == EscapeState.in_tag and ch2 == DJANGO_TAG_CLOSE) or (
            state == EscapeState.in_exp and ch2 == DJANGO_EXP_CLOSE
        ):
            state = EscapeState.normal
            stream.ptr += 2
            new_text.append(ch2)
        else:
            stream.ptr += 1
            new_text.append(HTML_CHARS.get(ch, ch) if state == EscapeState.normal else ch)

    return "".join(new_text)



import unittest

from hamlpy.parser.core import (
    ParseException,
    Stream,
    peek_indentation,
    read_line,
    read_number,
    read_quoted_string,
    read_symbol,
    read_whitespace,
    read_word,
)
from hamlpy.parser.utils import html_escape


class ParserTest(unittest.TestCase):
    def test_read_whitespace(self):
        stream = Stream(" \t foo  \n  bar   ")

        assert read_whitespace(stream) == " \t "
        assert stream.text[stream.ptr :] == "foo  \n  bar   "

        stream.ptr += 3  # skip over foo

        assert read_whitespace(stream) == "  "
        assert stream.text[stream.ptr :] == "\n  bar   "

        assert read_whitespace(stream, include_newlines=True) == "\n  "
        assert stream.text[stream.ptr :] == "bar   "

        stream.ptr += 3  # skip over bar

        assert read_whitespace(stream) == "   "
        assert stream.text[stream.ptr :] == ""

    def test_peek_indentation(self):
        assert peek_indentation(Stream("content")) == 0
        assert peek_indentation(Stream("  content")) == 2
        assert peek_indentation(Stream("\n")) is None
        assert peek_indentation(Stream("    \n")) is None

    def test_quoted_string(self):
        stream = Stream("'hello'---")
        assert read_quoted_string(stream) == "hello"
        assert stream.text[stream.ptr :] == "---"

        stream = Stream('"this don\'t \\"x\\" hmm" not in string')
        assert read_quoted_string(stream) == 'this don\'t "x" hmm'
        assert stream.text[stream.ptr :] == " not in string"

        self.assertRaises(ParseException, read_quoted_string, Stream('"no end quote...'))

    def test_read_line(self):
        stream = Stream("line1\n line2\n\nline4\n\n")
        assert read_line(stream) == "line1"
        assert read_line(stream) == " line2"
        assert read_line(stream) == ""
        assert read_line(stream) == "line4"
        assert read_line(stream) == ""
        assert read_line(stream) is None

        assert read_line(Stream("last line  ")) == "last line  "

    def test_read_number(self):
        stream = Stream('123"')
        assert read_number(stream) == "123"
        assert stream.text[stream.ptr :] == '"'

        stream = Stream("123.4xx")
        assert read_number(stream) == "123.4"
        assert stream.text[stream.ptr :] == "xx"

        stream = Stream("0.0001   ")
        assert read_number(stream) == "0.0001"
        assert stream.text[stream.ptr :] == "   "

    def test_read_symbol(self):
        stream = Stream("=> bar")
        assert read_symbol(stream, ["=>", ":"]) == "=>"
        assert stream.text[stream.ptr :] == " bar"

        self.assertRaises(ParseException, read_symbol, Stream("foo"), ["=>"])

    def test_read_word(self):
        stream = Stream("foo_bar")
        assert read_word(stream) == "foo_bar"
        assert stream.text[stream.ptr :] == ""

        stream = Stream("foo_bar ")
        assert read_word(stream) == "foo_bar"
        assert stream.text[stream.ptr :] == " "

        stream = Stream("ng-repeat(")
        assert read_word(stream) == "ng"
        assert stream.text[stream.ptr :] == "-repeat("

        stream = Stream("ng-repeat(")
        assert read_word(stream, ("-",)) == "ng-repeat"
        assert stream.text[stream.ptr :] == "("

        stream = Stream("これはテストです...")
        assert read_word(stream) == "これはテストです"
        assert stream.text[stream.ptr :] == "..."


class UtilsTest(unittest.TestCase):
    def test_html_escape(self):
        assert html_escape("") == ""
        assert html_escape("&<>\"'") == "&amp;&lt;&gt;&quot;&#39;"
        assert html_escape('{% trans "hello" %}') == '{% trans "hello" %}'
        assert html_escape('{{ foo|default:"hello" }}') == '{{ foo|default:"hello" }}'
        assert html_escape("{% }} & %}") == "{% }} & %}"

        result = html_escape('<>{% trans "hello" %}<>{{ foo|default:"hello" }}<>')
        assert result == '&lt;&gt;{% trans "hello" %}&lt;&gt;{{ foo|default:"hello" }}&lt;&gt;'



import unittest

from hamlpy.compiler import Compiler
from hamlpy.parser import filters
from hamlpy.parser.core import ParseException


class CompilerTest(unittest.TestCase):
    def test_tags(self):
        # tags can have xml namespaces
        self._test("%fb:tag\n  content", "<fb:tag>\n  content\n</fb:tag>")

        # tags can have dashes
        self._test("%ng-tag\n  content", "<ng-tag>\n  content\n</ng-tag>")

    def test_ids_and_classes(self):
        # id on tag
        self._test("%div#someId Some text", "<div id='someId'>Some text</div>")

        # id with non-ascii characters
        self._test("%div#これはテストです test", "<div id='これはテストです'>test</div>")

        # class on tag
        self._test("%div.someClass Some text", "<div class='someClass'>Some text</div>")

        # class can contain dash
        self._test(".header.span-24.last", "<div class='header span-24 last'></div>")

        # multiple classes
        self._test("%div.someClass.anotherClass Some text", "<div class='someClass anotherClass'>Some text</div>")

        # class can come before id
        self._test("%div.someClass#someId", "<div class='someClass' id='someId'></div>")

    def test_attribute_dictionaries(self):
        # attribute dictionaries
        self._test(
            "%html{'xmlns':'http://www.w3.org/1999/xhtml', 'xml:lang':'en', 'lang':'en'}",
            "<html xmlns='http://www.w3.org/1999/xhtml' xml:lang='en' lang='en'></html>",
        )

        # attribute whitespace is ignored
        self._test('%form{ id : "myform" }', "<form id='myform'></form>")

        # HTML style
        self._test('%form(foo=bar id="myform")', "<form foo='{{ bar }}' id='myform'></form>")

        # multiple dicts
        self._test('%a(a="b"){:c => "d"} Stuff', "<a a='b' c='d'>Stuff</a>")

        self._test_error("%a(b=)", 'Unexpected ")". @ "%a(b=)" <-')

    def test_boolean_attributes(self):
        self._test("%input{required}", "<input required>")
        self._test("%input{required, a: 'b'}", "<input required a='b'>")
        self._test("%input{a: 'b', required, b: 'c'}", "<input a='b' required b='c'>")
        self._test("%input{a: 'b', required}", "<input a='b' required>")
        self._test("%input{checked, required, visible}", "<input checked required visible>")
        self._test("%input(checked=true)", "<input checked>")
        self._test("%input(checked=true)", "<input checked='checked' />", options={"format": "xhtml"})

    def test_attribute_values_as_tuples_and_lists(self):
        # id attribute as tuple
        self._test("%div{'id':('itemType', '5')}", "<div id='itemType_5'></div>")

        # attributes as lists
        self._test(
            "%div{'id':['Article','1'], 'class':['article','entry','visible']} Booyaka",
            "<div class='article entry visible' id='Article_1'>Booyaka</div>",
        )
        self._test(
            "%div{'id': ('article', '3'), 'class': ('newest', 'urgent')} Content",
            "<div class='newest urgent' id='article_3'>Content</div>",
        )

    def test_comments(self):
        self._test("/ some comment", "<!-- some comment -->")

        # comments should hide child nodes
        self._test(
            """
-# My comment
  #my_div
    my text
test""",
            "test",
        )

        # conditional comments
        self._test("/[if IE] You use a shitty browser", "<!--[if IE]> You use a shitty browser<![endif]-->")
        self._test(
            "/[if IE]\n  %h1 You use a shitty browser",
            "<!--[if IE]>\n  <h1>You use a shitty browser</h1>\n<![endif]-->",
        )
        self._test("/[if lte IE 7]\n\ttest\n#test", "<!--[if lte IE 7]>\n\ttest\n<![endif]-->\n<div id='test'></div>")

    def test_django_variables(self):
        # whole-line variable with =
        self._test("= story.tease", "{{ story.tease }}")

        # element content variable with =
        self._test("%div= story.tease", "<div>{{ story.tease }}</div>")

        # embedded Django variables using #{...}
        self._test("#{greeting} #{name}, how are you #{date}?", "{{ greeting }} {{ name }}, how are you {{ date }}?")
        self._test("%h1 Hello, #{person.name}, how are you?", "<h1>Hello, {{ person.name }}, how are you?</h1>")

        # embedded Django variables using ={...} (not enabled by default)
        self._test("Hi ={name}, how are you?", "Hi ={name}, how are you?")
        self._test("Hi ={name}, how are you?", "Hi {{ name }}, how are you?", options={"django_inline_style": True})

        # variables can use Django filters
        self._test('#{value|center:"15"}', '{{ value|center:"15" }}')

        # variables can be used in attribute values
        self._test("%a{'b': '#{greeting} test'} blah", "<a b='{{ greeting }} test'>blah</a>")

        # including in the id or class
        self._test("%div{'id':'package_#{object.id}'}", "<div id='package_{{ object.id }}'></div>")
        self._test("%div{'class':'package_#{object.id}'}", "<div class='package_{{ object.id }}'></div>")

        # they can be escaped
        self._test(
            "%a{'b': '\\\\#{greeting} test', title: \"It can't be removed\"} blah",
            "<a b='#{greeting} test' title='It can&#39;t be removed'>blah</a>",
        )
        self._test(
            "%h1 Hello, \\={name}, how are you ={ date }?",
            "<h1>Hello, ={name}, how are you {{ date }}?</h1>",
            options={"django_inline_style": True},
        )
        self._test("\\#{name}, how are you?", "#{name}, how are you?")

    def test_django_tags(self):
        # if/else
        self._test(
            "- if something\n   %p hello\n- else\n   %p goodbye",
            "{% if something %}\n   <p>hello</p>\n{% else %}\n   <p>goodbye</p>\n{% endif %}",
        )

        # with
        self._test("- with thing1 as another\n  stuff", "{% with thing1 as another %}\n  stuff\n{% endwith %}")
        self._test(
            "- with context\n  hello\n- with other_context\n  goodbye",
            "{% with context %}\n  hello\n{% endwith %}\n{% with other_context %}\n  goodbye\n{% endwith %}",
        )

        # block
        self._test(
            "-block title\n   %p hello\n",
            "{% block title %}\n   <p>hello</p>\n{% endblock %}",
        )
        self._test(
            "-block title\n   %p hello\n",
            "{% block title %}\n   <p>hello</p>\n{% endblock title %}",
            options={"endblock_names": True},
        )

        # trans
        self._test('- trans "Hello"\n', '{% trans "Hello" %}')

        # blocktrans
        self._test(
            "- blocktrans with amount=num_cookies\n" "  There are #{ amount } cookies",
            "{% blocktrans with amount=num_cookies %}\n" "  There are {{ amount }} cookies\n" "{% endblocktrans %}",
        )
        self._test(
            "- blocktrans with amount=num_cookies\n"
            "  There is one cookie\n"
            "- plural\n"
            "  There are #{ amount } cookies",
            "{% blocktrans with amount=num_cookies %}\n"
            "  There is one cookie\n"
            "{% plural %}\n"
            "  There are {{ amount }} cookies\n"
            "{% endblocktrans %}",
        )

        # exception using a closing tag of a self-closing tag
        parser = Compiler()
        self.assertRaises(ParseException, parser.process, "- endfor")

    def test_plain_text(self):
        self._test("This should be plain text", "This should be plain text")
        self._test(
            "This should be plain text\n    This should be indented",
            "This should be plain text\n    This should be indented",
        )

        # native Django tags {% %} should be treated as plain text
        self._test("text   {%\n  trans ''\n%}", "text   {%\n  trans ''\n%}")
        self._test("text\n   {%\n  trans ''\n%}", "text\n   {%\n  trans ''\n%}")

    def test_plain_filter(self):
        # with indentation
        self._test(
            ":plain\n    -This should be plain text\n    .This should be more\n      This should be indented",
            "-This should be plain text\n.This should be more\n  This should be indented",
        )

        # with no children
        self._test(":plain\nNothing", "Nothing")

        # with escaped back slash
        self._test(":plain\n  \\Something", "\\Something")

        # with space after filter name
        self._test(":plain \n    -This should be plain text\n", "-This should be plain text")

    def test_preserve_filter(self):
        # with indentation
        self._test(
            ":preserve\n    -This should be plain text\n    .This should be more\n      This should be indented",
            "-This should be plain text&#x000A;.This should be more&#x000A;  This should be indented",
        )

        # with no children
        self._test(":preserve\nNothing", "Nothing")

        # with escaped back slash
        self._test(":preserve\n  \\Something", "\\Something")

    def test_python_filter(self):
        self._test(":python\n", "")  # empty
        self._test(
            ':python\n   for i in range(0, 5): print("<p>item \\%s</p>" % i)',
            "<p>item \\0</p>\n<p>item \\1</p>\n<p>item \\2</p>\n<p>item \\3</p>\n<p>item \\4</p>",
        )

        self._test_error(":python\n   print(10 / 0)", "Error whilst executing python filter node", ZeroDivisionError)

    def test_pygments_filter(self):
        self._test(":highlight\n", "")  # empty
        self._test(
            ":highlight\n  print(1)\n", '<div class="highlight"><pre><span></span>print(1)\n</pre></div>'
        )  # noqa

        filters._pygments_available = False

        self._test_error(":highlight\n  print(1)\n", "Pygments is not available")

        filters._pygments_available = True

    def test_markdown_filter(self):
        self._test(":markdown\n", "")  # empty
        self._test(":markdown\n  *Title*\n", "<p><em>Title</em></p>")

        filters._markdown_available = False

        self._test_error(":markdown\n  *Title*\n", "Markdown is not available")

        filters._markdown_available = True

    def test_invalid_filter(self):
        self._test_error(":nosuchfilter\n", "No such filter: nosuchfilter")

    def test_doctypes(self):
        self._test("!!!", "<!DOCTYPE html>", options={"format": "html5"})
        self._test("!!! 5", "<!DOCTYPE html>", options={"format": "xhtml"})
        self._test("!!! 5", "<!DOCTYPE html>")
        self._test(
            "!!! strict",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!! frameset",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!! mobile",
            '<!DOCTYPE html PUBLIC "-//WAPFORUM//DTD XHTML Mobile 1.2//EN" "http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!! basic",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML Basic 1.1//EN" "http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!! transitional",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!!",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',  # noqa
            options={"format": "xhtml"},
        )
        self._test(
            "!!! strict",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">',  # noqa
            options={"format": "html4"},
        )
        self._test(
            "!!! frameset",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">',  # noqa
            options={"format": "html4"},
        )
        self._test(
            "!!! transitional",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">',  # noqa
            options={"format": "html4"},
        )
        self._test(
            "!!!",
            '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">',  # noqa
            options={"format": "html4"},
        )

        self._test("!!! XML", "", options={"format": "html4"})
        self._test("!!! XML iso-8589", "<?xml version='1.0' encoding='iso-8589' ?>", options={"format": "xhtml"})

    def test_attr_wrapper(self):
        self._test(
            "%p{ :strange => 'attrs'}", "<p strange=*attrs*></p>", options={"attr_wrapper": "*", "escape_attrs": True}
        )
        self._test(
            "%p{ :escaped => 'quo\"te'}",
            '<p escaped="quo&quot;te"></p>',
            options={"attr_wrapper": '"', "escape_attrs": True},
        )
        self._test(
            "%p{ :escaped => 'quo\\'te'}",
            '<p escaped="quo&#39;te"></p>',
            options={"attr_wrapper": '"', "escape_attrs": True},
        )
        self._test(
            "%p{ :escaped => 'q\\'uo\"te'}",
            '<p escaped="q&#39;uo&quot;te"></p>',
            options={"attr_wrapper": '"', "escape_attrs": True},
        )
        self._test(
            "!!! XML",
            '<?xml version="1.0" encoding="utf-8" ?>',
            options={"attr_wrapper": '"', "format": "xhtml", "escape_attrs": True},
        )

        # smart quotes...

        self._test(
            """%a(onclick="alert('hi')")""",
            """<a onclick="alert('hi')"></a>""",
            options={"attr_wrapper": '"', "smart_quotes": False},
        )
        self._test(
            """%a(onclick="alert('hi')")""",
            """<a onclick="alert('hi')"></a>""",
            options={"attr_wrapper": '"', "smart_quotes": True},
        )
        self._test(
            """%a(onclick="alert('hi')")""",
            """<a onclick='alert("hi")'></a>""",
            options={"attr_wrapper": "'", "smart_quotes": True},
        )
        self._test(
            """%a(onclick='alert("hi {% trans "world" %}")')""",
            """<a onclick="alert('hi {% trans "world" %}')"></a>""",
            options={"attr_wrapper": '"', "smart_quotes": True},
        )

    def test_attr_escaping(self):
        self._test(
            """#foo{:class => '<?php echo "&quot;" ?>'}""",
            """<div class='<?php echo "&quot;" ?>' id='foo'></div>""",
            options={"escape_attrs": False},
        )
        self._test(
            """#foo{:class => '"&lt;&gt;&amp;"'}""",
            """<div class='&quot;&amp;lt;&amp;gt;&amp;amp;&quot;' id='foo'></div>""",
            options={"escape_attrs": True},
        )
        self._test(
            """#foo{:class => '{% trans "Hello" %}'}""",
            """<div class='{% trans "Hello" %}' id='foo'></div>""",
            options={"escape_attrs": True},
        )

    def test_node_escaping(self):
        self._test("\\= Escaped", "= Escaped")
        self._test("\\%}", "%}")
        self._test("  \\:python", "  :python")

    def test_utf8(self):
        self._test(
            "%a{'href':'', 'title':'링크(Korean)'} Some Link", "<a href='' title='\ub9c1\ud06c(Korean)'>Some Link</a>"
        )

    def test_custom_filter(self):
        def upper(text, options):
            return text.upper()

        filters.register_filter("upper", upper)

        self._test(":upper\n  welcome", "WELCOME")

    def _test(self, haml, expected_html, options=None):
        if not options:
            options = {"escape_attrs": True}

        compiler = Compiler(options)
        result = compiler.process(haml)

        result = result.rstrip("\n")  # ignore trailing new lines

        assert result == expected_html

    def _test_error(self, haml, expected_message, expected_cause=None, compiler_options=None):
        compiler = Compiler(compiler_options)

        try:
            compiler.process(haml)
            self.fail("Expected exception to be raised")
        except Exception as e:
            self.assertIsInstance(e, ParseException)
            assert str(e) == expected_message

            if expected_cause:
                assert type(e.__cause__) == expected_cause



import textwrap

from .core import ParseException, TreeNode, peek_indentation, read_line, read_whitespace
from .elements import read_element
from .filters import get_filter

XHTML_DOCTYPES = {
    "1.1": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',  # noqa
    "strict": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',  # noqa
    "frameset": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">',  # noqa
    "mobile": '<!DOCTYPE html PUBLIC "-//WAPFORUM//DTD XHTML Mobile 1.2//EN" "http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd">',  # noqa
    "rdfa": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML+RDFa 1.0//EN" "http://www.w3.org/MarkUp/DTD/xhtml-rdfa-1.dtd">',  # noqa
    "basic": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML Basic 1.1//EN" "http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd">',  # noqa
    "": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',  # noqa
}

HTML4_DOCTYPES = {
    "strict": '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">',
    "frameset": '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">',
    "": '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">',
}


DOCTYPE_PREFIX = "!!!"
ELEMENT_PREFIXES = ("%", "#", ".")
HTML_COMMENT_PREFIX = "/"
CONDITIONAL_COMMENT_PREFIX = "/["
HAML_COMMENT_PREFIX = "-#"
VARIABLE_PREFIX = "="
TAG_PREFIX = "-"
FILTER_PREFIX = ":"

HAML_ESCAPE = "\\"


def read_node(stream, prev, compiler):
    """
    Reads a node, returning either the node or None if we've reached the end of the input
    """
    while True:
        indent = read_whitespace(stream)

        if stream.ptr >= stream.length:
            return None

        # convert indent to be all of the first character
        if indent:
            indent = indent[0] * len(indent)

        # empty lines are recorded as newlines on previous node
        if stream.text[stream.ptr] == "\n":
            if prev:
                prev.newlines += 1
            stream.ptr += 1
            continue

        # parse filter node
        if stream.text[stream.ptr] == FILTER_PREFIX:
            return read_filter_node(stream, indent, compiler)

        # peek ahead to so we don't try to parse an element from a variable node starting #{ or a Django tag ending %}
        if stream.text[stream.ptr] in ELEMENT_PREFIXES and stream.text[stream.ptr : stream.ptr + 2] not in (
            "#{",
            "%}",
        ):
            element = read_element(stream, compiler)
            return ElementNode(element, indent, compiler)

        # all other nodes are single line
        line = read_line(stream)

        inline_var_regex, escaped_var_regex = compiler.inline_variable_regexes

        if inline_var_regex.match(line) or escaped_var_regex.match(line):
            return PlaintextNode(line, indent, compiler)

        if line[0] == HAML_ESCAPE:
            return PlaintextNode(line, indent, compiler)

        if line.startswith(DOCTYPE_PREFIX):
            return DoctypeNode(line, indent, compiler)

        if line.startswith(CONDITIONAL_COMMENT_PREFIX):
            return ConditionalCommentNode(line, indent, compiler)

        if line[0] == HTML_COMMENT_PREFIX:
            return CommentNode(line, indent, compiler)

        if line.startswith(HAML_COMMENT_PREFIX):
            return HamlCommentNode(line, indent, compiler)

        if line[0] == VARIABLE_PREFIX:
            return VariableNode(line, indent, compiler)

        if line[0] == TAG_PREFIX:
            return TagNode(line, indent, compiler)

        return PlaintextNode(line, indent, compiler)


def read_filter_node(stream, indent, compiler):
    """
    Reads a filter node including its indented content, e.g. :plain
    """
    assert stream.text[stream.ptr] == FILTER_PREFIX

    stream.ptr += 1  # consume the initial colon
    name = read_line(stream)
    content_lines = []

    # read lines below with higher indentation as this filter's content
    while stream.ptr < stream.length:
        line_indentation = peek_indentation(stream)

        if line_indentation is not None and line_indentation <= len(indent):
            break

        line = read_line(stream)

        # don't preserve whitespace on empty lines
        if line.isspace():
            line = ""

        content_lines.append(line)

    return FilterNode(name.rstrip(), "\n".join(content_lines), indent, compiler)


class Node(TreeNode):
    """
    Base class of all nodes
    """

    def __init__(self, indent, compiler):
        super(Node, self).__init__()

        if indent is not None:
            self.indent = indent
            self.indentation = len(indent)
        else:
            self.indent = None
            self.indentation = -1

        self.compiler = compiler

        self.newlines = 0  # number of empty lines to render after node
        self.before = ""  # rendered text at start of node, e.g. "<p>\n"
        self.after = ""  # rendered text at end of node, e.g. "\n</p>"

    @classmethod
    def create_root(cls, compiler):
        return cls(None, compiler)

    def render(self):
        # Render (sets self.before and self.after)
        self._render_children()
        # Post-render (nodes can modify the rendered text of other nodes)
        self._post_render()
        # Generate HTML
        return self._generate_html()

    def render_newlines(self):
        return "\n" * (self.newlines + 1)

    def _render_children(self):
        for child in self.children:
            child._render()

    def _post_render(self):
        for child in self.children:
            child._post_render()

    def _generate_html(self):
        output = [self.before]
        for child in self.children:
            output.append(child.before)
            output += [gc._generate_html() for gc in child.children]
            output.append(child.after)
        output.append(self.after)
        return "".join(output)

    def replace_inline_variables(self, content):
        inline_var_regex, escaped_var_regex = self.compiler.inline_variable_regexes

        content = inline_var_regex.sub(r"{{ \2 }}", content)
        content = escaped_var_regex.sub(r"\1", content)
        return content

    def add_node(self, node):
        if self._should_go_inside_last_node(node):
            self.children[-1].add_node(node)
        else:
            self.add_child(node)

    def _should_go_inside_last_node(self, node):
        return len(self.children) > 0 and (
            node.indentation > self.children[-1].indentation
            or (node.indentation == self.children[-1].indentation and self.children[-1].should_contain(node))
        )

    def should_contain(self, node):
        return False

    def debug_tree(self):  # pragma: no cover
        return "\n".join(self._debug_tree([self]))

    def _debug_tree(self, nodes):  # pragma: no cover
        output = []
        for n in nodes:
            output.append("%s%s" % (" " * (n.indentation + 2), n))
            if n.children:
                output += self._debug_tree(n.children)
        return output

    def __repr__(self):  # pragma: no cover
        return "%s" % type(self).__name__


class LineNode(Node):
    """
    Base class of nodes which are a single line of Haml
    """

    def __init__(self, line, indent, compiler):
        super(LineNode, self).__init__(indent, compiler)

        self.haml = line.rstrip()

    def __repr__(self):  # pragma: no cover
        return "%s(indent=%d, newlines=%d): %s" % (type(self).__name__, self.indentation, self.newlines, self.haml)


class PlaintextNode(LineNode):
    """
    Node that is not modified or processed when rendering
    """

    def _render(self):
        text = self.replace_inline_variables(self.haml)

        # remove escape character
        if text and text[0] == HAML_ESCAPE:
            text = text.replace(HAML_ESCAPE, "", 1)

        self.before = "%s%s" % (self.indent, text)
        if self.children:
            self.before += self.render_newlines()
        else:
            self.after = self.render_newlines()

        self._render_children()


class ElementNode(Node):
    """
    An HTML tag node, e.g. %span
    """

    def __init__(self, element, indent, compiler):
        super(ElementNode, self).__init__(indent, compiler)

        self.element = element

    def _render(self):
        self.before = self._render_before(self.element)
        self.after = self._render_after(self.element)
        self._render_children()

    def _render_before(self, element):
        """
        Render opening tag and inline content
        """
        start = ["%s<%s" % (self.indent, element.tag)]

        attributes = element.render_attributes(self.compiler.options)
        if attributes:
            start.append(" " + self.replace_inline_variables(attributes))

        content = self._render_inline_content(self.element.inline_content)

        if element.nuke_inner_whitespace and content:
            content = content.strip()

        if element.self_close and not content:
            start.append(">" if self.compiler.options.html else " />")
        elif content:
            start.append(">%s" % content)
        elif self.children:
            start.append(">%s" % (self.render_newlines()))
        else:
            start.append(">")
        return "".join(start)

    def _render_after(self, element):
        """
        Render closing tag
        """
        if element.inline_content:
            return "</%s>%s" % (element.tag, self.render_newlines())
        elif element.self_close:
            return self.render_newlines()
        elif self.children:
            return "%s</%s>\n" % (self.indent, element.tag)
        else:
            return "</%s>\n" % element.tag

    def _post_render(self):
        # inner whitespace removal
        if self.element.nuke_inner_whitespace:
            self.before = self.before.rstrip()
            self.after = self.after.lstrip()

            if self.children:
                node = self
                if node.children:
                    node.children[0].before = node.children[0].before.lstrip()

                if node.children:
                    node.children[-1].after = node.children[-1].after.rstrip()

        # outer whitespace removal
        if self.element.nuke_outer_whitespace:
            left_sibling = self.left_sibling()
            if left_sibling:
                # If node has left sibling, strip whitespace after left sibling
                left_sibling.after = left_sibling.after.rstrip()
                left_sibling.newlines = 0
            else:
                # If not, whitespace comes from it's parent node,
                # so strip whitespace before the node
                self.parent.before = self.parent.before.rstrip()
                self.parent.newlines = 0

            self.before = self.before.lstrip()
            self.after = self.after.rstrip()

            right_sibling = self.right_sibling()
            if right_sibling:
                right_sibling.before = right_sibling.before.lstrip()
            else:
                self.parent.after = self.parent.after.lstrip()
                self.parent.newlines = 0

        super(ElementNode, self)._post_render()

    def _render_inline_content(self, inline_content):
        if inline_content is None or len(inline_content) == 0:
            return None

        if self.element.django_variable:
            content = "{{ " + inline_content.strip() + " }}"
            return content
        else:
            return self.replace_inline_variables(inline_content)


class CommentNode(LineNode):
    """
    An HTML comment node, e.g. / This is a comment
    """

    def _render(self):
        self.after = "-->\n"
        if self.children:
            self.before = "<!-- %s" % (self.render_newlines())
            self._render_children()
        else:
            self.before = "<!-- %s " % (self.haml.lstrip(HTML_COMMENT_PREFIX).strip())


class ConditionalCommentNode(LineNode):
    """
    An HTML conditional comment node, e.g. /[if IE]
    """

    def _render(self):
        conditional = self.haml[1 : self.haml.index("]") + 1]

        if self.children:
            self.before = "<!--%s>\n" % conditional
        else:
            content = self.haml[len(CONDITIONAL_COMMENT_PREFIX) + len(conditional) - 1 :]
            self.before = "<!--%s>%s" % (conditional, content)

        self.after = "<![endif]-->\n"
        self._render_children()


class DoctypeNode(LineNode):
    """
    An XML doctype node, e.g. !!! 5
    """

    def _render(self):
        doctype = self.haml.lstrip(DOCTYPE_PREFIX).strip().lower()

        self.before = self.get_header(doctype, self.compiler.options)
        self.after = self.render_newlines()

    def get_header(self, doctype, options):
        if doctype.startswith("xml"):
            if options.html:
                return ""
            parts = doctype.split()
            encoding = parts[1] if len(parts) > 1 else "utf-8"
            return "<?xml version=%s1.0%s encoding=%s%s%s ?>" % (
                options.attr_wrapper,
                options.attr_wrapper,
                options.attr_wrapper,
                encoding,
                options.attr_wrapper,
            )
        elif options.html5:
            return "<!DOCTYPE html>"
        elif options.xhtml:
            if doctype == "5":
                return "<!DOCTYPE html>"
            else:
                return XHTML_DOCTYPES.get(doctype, XHTML_DOCTYPES[""])
        else:
            return HTML4_DOCTYPES.get(doctype, HTML4_DOCTYPES[""])


class HamlCommentNode(LineNode):
    """
    A Haml comment node, e.g. -# This is a comment
    """

    def _render(self):
        self.after = self.render_newlines()[1:]

    def _post_render(self):
        pass


class VariableNode(LineNode):
    """
    A Django variable node, e.g. =person.name
    """

    def __init__(self, haml, indent, compiler):
        super(VariableNode, self).__init__(haml, indent, compiler)

    def _render(self):
        tag_content = self.haml.lstrip(VARIABLE_PREFIX)
        self.before = "%s{{ %s }}" % (self.indent, tag_content.strip())
        self.after = self.render_newlines()

    def _post_render(self):
        pass


class TagNode(LineNode):
    """
    A Django/Jinja server-side tag node, e.g. -block
    """

    def __init__(self, haml, indent, compiler):
        super(TagNode, self).__init__(haml, indent, compiler)

        self.tag_statement = self.haml.lstrip(TAG_PREFIX).strip()
        self.tag_name = self.tag_statement.split(" ")[0]

        if self.tag_name in self.compiler.self_closing_tags.values():
            raise ParseException("Unexpected closing tag for self-closing tag %s" % self.tag_name)

    def _render(self):
        self.before = "%s{%% %s %%}" % (self.indent, self.tag_statement)

        closing_tag = self.compiler.self_closing_tags.get(self.tag_name)

        if closing_tag:
            if self.tag_name == "block" and self.compiler.options.endblock_names:
                block_name = self.tag_statement.split(" ")[1]
                closing_tag += " " + block_name

            self.before += self.render_newlines()
            self.after = "%s{%% %s %%}%s" % (self.indent, closing_tag, self.render_newlines())
        else:
            if self.children:
                self.before += self.render_newlines()
            else:
                self.after = self.render_newlines()
        self._render_children()

    def should_contain(self, node):
        return isinstance(node, TagNode) and node.tag_name in self.compiler.tags_may_contain.get(self.tag_name, "")


class FilterNode(Node):
    """
    A type filter, e.g. :javascript
    """

    def __init__(self, filter_name, content, indent, compiler):
        super(FilterNode, self).__init__(indent, compiler)

        self.filter_name = filter_name
        self.content = content

    def _render(self):
        content = textwrap.dedent(self.content)

        filter_func = get_filter(self.filter_name)
        content = filter_func(content, self.compiler.options)

        content = self.indent + content.replace("\n", "\n" + self.indent)

        self.before = content
        self.after = self.render_newlines() if self.content else ""

    def _post_render(self):
        pass

    def __repr__(self):  # pragma: no cover
        return "%s(indent=%d, newlines=%d, filter=%s): %s" % (
            type(self).__name__,
            self.indentation,
            self.newlines,
            self.filter_name,
            self.content,
        )



from collections import OrderedDict

from .attributes import read_attribute_dict
from .core import read_line, read_word

# non-word characters that we allow in tag names, ids and classes
TAG_EXTRA_CHARS = ("-",)

# non-word characters that we allow in ids and classes
ID_OR_CLASS_EXTRA_CHARS = ("-", ":")


def read_tag(stream):
    """
    Reads an element tag, e.g. span, ng-repeat, cs:dropdown
    """
    part1 = read_word(stream, TAG_EXTRA_CHARS)

    if stream.ptr < stream.length and stream.text[stream.ptr] == ":":
        stream.ptr += 1
        part2 = read_word(stream, TAG_EXTRA_CHARS)
    else:
        part2 = None

    return (part1 + ":" + part2) if part2 else part1


def read_element(stream, compiler):
    """
    Reads an element, e.g. %span, #banner{style:"width: 100px"}, .ng-hide(foo=1)
    """
    assert stream.text[stream.ptr] in ("%", ".", "#")

    tag = None
    empty_class = False

    if stream.text[stream.ptr] == "%":
        stream.ptr += 1
        tag = read_tag(stream)

    elif stream.text[stream.ptr] == ".":
        # Element may start with a period representing an unidentified div rather than a CSS class. In this case it
        # can't have other classes or ids, e.g. .{foo:"bar"}
        next_ch = stream.text[stream.ptr + 1] if stream.ptr < stream.length - 1 else None
        if not (next_ch.isalnum() or next_ch == "_" or next_ch in ID_OR_CLASS_EXTRA_CHARS):
            stream.ptr += 1
            empty_class = True

    _id = None
    classes = []

    if not empty_class:
        while stream.ptr < stream.length and stream.text[stream.ptr] in ("#", "."):
            is_id = stream.text[stream.ptr] == "#"
            stream.ptr += 1

            id_or_class = read_word(stream, ID_OR_CLASS_EXTRA_CHARS)
            if is_id:
                _id = id_or_class
            else:
                classes.append(id_or_class)

    attributes = OrderedDict()
    while stream.ptr < stream.length and stream.text[stream.ptr] in ("{", "("):
        attributes.update(read_attribute_dict(stream, compiler))

    if stream.ptr < stream.length and stream.text[stream.ptr] == ">":
        stream.ptr += 1
        nuke_outer_ws = True
    else:
        nuke_outer_ws = False

    if stream.ptr < stream.length and stream.text[stream.ptr] == "<":
        stream.ptr += 1
        nuke_inner_ws = True
    else:
        nuke_inner_ws = False

    if stream.ptr < stream.length and stream.text[stream.ptr] == "/":
        stream.ptr += 1
        self_close = True
    else:
        self_close = tag in Element.SELF_CLOSING

    if stream.ptr < stream.length and stream.text[stream.ptr] == "=":
        stream.ptr += 1
        django_variable = True
    else:
        django_variable = False

    if stream.ptr < stream.length:
        inline = read_line(stream)
        if inline is not None:
            inline = inline.strip()
    else:
        inline = None

    return Element(tag, _id, classes, attributes, nuke_outer_ws, nuke_inner_ws, self_close, django_variable, inline)


class Element(object):
    """
    An HTML element with an id, classes, attributes etc
    """

    SELF_CLOSING = (
        "meta",
        "img",
        "link",
        "br",
        "hr",
        "input",
        "source",
        "track",
        "area",
        "base",
        "col",
        "command",
        "embed",
        "keygen",
        "param",
        "wbr",
    )

    DEFAULT_TAG = "div"
    ESCAPED = {'"': "&quot;", "'": "&#39;"}

    def __init__(
        self,
        tag,
        _id,
        classes,
        attributes,
        nuke_outer_whitespace,
        nuke_inner_whitespace,
        self_close,
        django_variable,
        inline_content,
    ):
        self.tag = tag or self.DEFAULT_TAG
        self.attributes = attributes
        self.nuke_inner_whitespace = nuke_inner_whitespace
        self.nuke_outer_whitespace = nuke_outer_whitespace
        self.self_close = self_close
        self.django_variable = django_variable
        self.inline_content = inline_content

        # merge ids from the attribute dictionary
        ids = [_id] if _id else []
        id_from_attrs = attributes.get("id")
        if isinstance(id_from_attrs, (tuple, list)):
            ids += id_from_attrs
        elif isinstance(id_from_attrs, str):
            ids += [id_from_attrs]

        # merge ids to a single value with _ separators
        self.id = "_".join(ids) if ids else None

        # merge classes from the attribute dictionary
        class_from_attrs = attributes.get("class", [])
        if not isinstance(class_from_attrs, (tuple, list)):
            class_from_attrs = [class_from_attrs]

        self.classes = class_from_attrs + classes

    def render_attributes(self, options):
        def attr_wrap(val):
            return "%s%s%s" % (options.attr_wrapper, val, options.attr_wrapper)

        rendered = []

        for name, value in self.attributes.items():
            if name in ("id", "class") or value in (None, False):
                # this line isn't recorded in coverage because it gets optimized away (http://bugs.python.org/issue2506)
                continue  # pragma: no cover

            if value is True:  # boolean attribute
                if options.xhtml:
                    rendered.append("%s=%s" % (name, attr_wrap(name)))
                else:
                    rendered.append(name)
            else:
                value = self._escape_attribute_quotes(value, options.attr_wrapper, options.smart_quotes)
                rendered.append("%s=%s" % (name, attr_wrap(value)))

        if len(self.classes) > 0:
            rendered.append("class=%s" % attr_wrap(" ".join(self.classes)))

        if self.id:
            rendered.append("id=%s" % attr_wrap(self.id))

        return " ".join(rendered)

    @classmethod
    def _escape_attribute_quotes(cls, v, attr_wrapper, smart_quotes=False):
        """
        Escapes quotes, except those inside a Django tag
        """
        escaped = []
        inside_tag = False
        for i, _ in enumerate(v):
            if v[i : i + 2] == "{%":
                inside_tag = True
            elif v[i : i + 2] == "%}":
                inside_tag = False

            if v[i] == attr_wrapper and not inside_tag:
                if smart_quotes and attr_wrapper in ('"', "'"):
                    repl = '"' if v[i] == "'" else "'"
                else:
                    repl = cls.ESCAPED[attr_wrapper]

                escaped.append(repl)
            else:
                escaped.append(v[i])

        return "".join(escaped)



from django.apps import AppConfig


class Config(AppConfig):
    name = "hamlpy"

    def ready(self):
        # patch Django's templatize method
        from .template import templatize  # noqa


default_app_config = "hamlpy.Config"



from django.test import SimpleTestCase


class TemplatizeTest(SimpleTestCase):
    def test_templatize(self):
        from django.utils.translation import templatize

        # test regular Django tags
        output = templatize('{% trans "Foo" %}{% blocktrans %}\nBar\n{% endblocktrans %}', origin="test.html")
        self.assertRegex(output, r"gettext\(u?'Foo'\)")
        self.assertRegex(output, r"gettext\(u?'\\nBar\\n'\)")

        # test Haml tags with HTML origin
        output = templatize('- trans "Foo"\n- blocktrans\n  Bar\n', origin="test.haml")
        self.assertRegex(output, r"gettext\(u?'Foo'\)")
        self.assertRegex(output, r"gettext\(u?'\\n  Bar\\n'\)")

        # test Haml tags and HTML origin
        self.assertNotIn("gettext", templatize('- trans "Foo"\n- blocktrans\n  Bar\n', origin="test.html"))

        # test Haml tags and no origin
        self.assertNotIn("gettext", templatize('- trans "Foo"\n- blocktrans\n  Bar\n'))



import sys
from io import StringIO

from django.conf import settings

from .core import ParseException
from .utils import html_escape

"""
Core HamlPy filters.

The implementation of these should match https://github.com/haml/haml/blob/master/lib/haml/filters.rb as closely as
possible. Where we differ is that we don't compile Stylus, Coffeescript etc into CSS or Javascript - but place the
content into suitable <style> and <script> that can be transformed later by something like django-compressor.
"""


# Pygments and Markdown are optional dependencies which may or may not be available
try:
    import pygments
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import PythonLexer, guess_lexer
    from pygments.util import ClassNotFound

    _pygments_available = True
except ImportError:  # pragma: no cover
    _pygments_available = False

try:
    from markdown import markdown as markdown_lib

    _markdown_available = True
except ImportError:  # pragma: no cover
    _markdown_available = False

# ----------------------------------------------------------------------------------
# Core filters
# ----------------------------------------------------------------------------------


def plain(text, options):
    return text


def preserve(text, options):
    text = text.rstrip()
    text = text.replace("\n", "&#x000A;")
    return text.replace("\r", "")


def escaped(text, options):
    return html_escape(text)


def cdata(text, options):
    text = "\n" + text.rstrip()
    text = text.replace("\n", "\n    ")
    return "<![CDATA[%s\n]]>" % text


def css(text, options):
    return style_filter(text, "text/css", options)


def stylus(text, options):
    return style_filter(text, "text/stylus", options)


def less(text, options):
    return style_filter(text, "text/less", options)


def sass(text, options):
    return style_filter(text, "text/sass", options)


def javascript(text, options):
    return script_filter(text, "text/javascript", "//", options)


def coffee(text, options):
    return script_filter(text, "text/coffeescript", "#", options)


def markdown(content, options):
    if not _markdown_available:
        raise ParseException("Markdown is not available")

    md_extensions = getattr(settings, "HAMLPY_MARKDOWN_EXTENSIONS", [])
    return markdown_lib(content, extensions=md_extensions)


def highlight(content, options):
    if not _pygments_available:
        raise ParseException("Pygments is not available")

    if content:
        # let Pygments try to guess syntax but default to Python
        try:
            lexer = guess_lexer(content)
        except ClassNotFound:
            lexer = PythonLexer()

        return pygments.highlight(content, lexer, HtmlFormatter())
    else:
        return ""


def python(content, options):
    if content:
        compiled_code = compile(content, "", "exec")
        output_buffer = StringIO()
        sys.stdout = output_buffer

        try:
            exec(compiled_code)
        except Exception as e:
            raise ParseException("Error whilst executing python filter node") from e
        finally:
            # restore the original stdout
            sys.stdout = sys.__stdout__

        return output_buffer.getvalue()
    else:
        return ""


# ----------------------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------------------


def style_filter(text, mime_type, options):
    indent = "    " if options.cdata else "  "
    text = text.rstrip().replace("\n", "\n" + indent)
    type_attr = " type=%(attr_wrapper)s%(mime_type)s%(attr_wrapper)s" % {
        "attr_wrapper": options.attr_wrapper,
        "mime_type": mime_type,
    }
    before, after = ("  /*<![CDATA[*/\n", "  /*]]>*/\n") if options.cdata else ("", "")

    return "<style%s>\n%s%s%s\n%s</style>" % (type_attr, before, indent, text, after)


def script_filter(text, mime_type, comment, options):
    indent = "    " if options.cdata else "  "
    text = text.rstrip().replace("\n", "\n" + indent)
    type_attr = " type=%(attr_wrapper)s%(mime_type)s%(attr_wrapper)s" % {
        "attr_wrapper": options.attr_wrapper,
        "mime_type": mime_type,
    }
    before, after = ("  %s<![CDATA[\n" % comment, "  %s]]>\n" % comment) if options.cdata else ("", "")

    return "<script%s>\n%s%s%s\n%s</script>" % (type_attr, before, indent, text, after)


# ----------------------------------------------------------------------------------
# Filter registration
# ----------------------------------------------------------------------------------

FILTERS = {
    "plain": plain,
    "preserve": preserve,
    "escaped": escaped,
    "cdata": cdata,
    "css": css,
    "stylus": stylus,
    "less": less,
    "sass": sass,
    "javascript": javascript,
    "coffee": coffee,
    "coffeescript": coffee,
    "markdown": markdown,
    "highlight": highlight,
    "python": python,
}


def register_filter(name, callback):
    FILTERS[name] = callback


def get_filter(name):
    if name not in FILTERS:
        raise ParseException("No such filter: " + name)

    return FILTERS.get(name)



DEBUG = True

DATABASES = {}

INSTALLED_APPS = ("hamlpy",)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": ["hamlpy/test/templates"],
        "OPTIONS": {
            "loaders": [
                "hamlpy.template.loaders.HamlPyFilesystemLoader",
                "hamlpy.template.loaders.HamlPyAppDirectoriesLoader",
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ],
            "debug": True,
        },
    }
]

SECRET_KEY = "tots top secret"



import unittest

from hamlpy.compiler import Compiler
from hamlpy.parser import nodes
from hamlpy.parser.core import Stream
from hamlpy.parser.nodes import read_filter_node, read_node


class NodeTest(unittest.TestCase):
    def test_read_node(self):
        assert isinstance(self._read_node("%div"), nodes.ElementNode)
        assert isinstance(self._read_node("   %html"), nodes.ElementNode)
        assert isinstance(self._read_node(".className"), nodes.ElementNode)
        assert isinstance(self._read_node("   .className"), nodes.ElementNode)
        assert isinstance(self._read_node("#idName"), nodes.ElementNode)
        assert isinstance(self._read_node("   #idName"), nodes.ElementNode)
        assert isinstance(self._read_node("/ some Comment"), nodes.CommentNode)
        assert isinstance(self._read_node("     / some Comment"), nodes.CommentNode)
        assert isinstance(self._read_node("just some random text"), nodes.PlaintextNode)
        assert isinstance(self._read_node("   more random text"), nodes.PlaintextNode)
        assert isinstance(self._read_node("-# This is a haml comment"), nodes.HamlCommentNode)
        assert isinstance(self._read_node("= some.variable"), nodes.VariableNode)
        assert isinstance(self._read_node("- for something in somethings"), nodes.TagNode)
        assert isinstance(self._read_node("\\= some.variable"), nodes.PlaintextNode)
        assert isinstance(self._read_node("    \\= some.variable"), nodes.PlaintextNode)
        assert isinstance(self._read_node("/[if IE 5]"), nodes.ConditionalCommentNode)
        assert isinstance(self._read_node(":plain"), nodes.FilterNode)
        assert isinstance(self._read_node("   :css\n"), nodes.FilterNode)
        assert isinstance(self._read_node(":stylus"), nodes.FilterNode)
        assert isinstance(self._read_node(":javascript"), nodes.FilterNode)
        assert isinstance(self._read_node(":coffee"), nodes.FilterNode)

    def test_read_filter_node(self):
        stream = Stream(':python\n  print("hello")\n')
        node = read_filter_node(stream, "", Compiler())
        assert node.filter_name == "python"
        assert node.content == '  print("hello")'
        assert stream.text[stream.ptr :] == ""

        stream = Stream(":javascript\n    var i = 0;\n  var j = 1;\n%span")
        node = read_filter_node(stream, "", Compiler())
        assert node.filter_name == "javascript"
        assert node.content == "    var i = 0;\n  var j = 1;"
        assert stream.text[stream.ptr :] == "%span"

    def test_calculates_indentation_properly(self):
        no_indentation = self._read_node("%div")
        assert no_indentation.indentation == 0

        three_indentation = self._read_node("   %div")
        assert three_indentation.indentation == 3

        six_indentation = self._read_node("      %div")
        assert six_indentation.indentation == 6

    def test_indents_tabs_properly(self):
        no_indentation = self._read_node("%div")
        assert no_indentation.indent == ""

        one_tab = self._read_node("	%div")
        assert one_tab.indent == "\t"

        one_space = self._read_node(" %div")
        assert one_space.indent == " "

        three_tabs = self._read_node("			%div")
        assert three_tabs.indent == "\t\t\t"

        tab_space = self._read_node("	 %div")
        assert tab_space.indent == "\t\t"

        space_tab = self._read_node(" 	%div")
        assert space_tab.indent == "  "

    def test_lines_are_always_stripped_of_whitespace(self):
        some_space = self._read_node("   text")
        assert some_space.haml == "text"

        lots_of_space = self._read_node("      text    ")
        assert lots_of_space.haml == "text"

    def test_inserts_nodes_into_proper_tree_depth(self):
        no_indentation_node = self._read_node("%div")
        one_indentation_node = self._read_node(" %div")
        two_indentation_node = self._read_node("  %div")
        another_one_indentation_node = self._read_node(" %div")

        no_indentation_node.add_node(one_indentation_node)
        no_indentation_node.add_node(two_indentation_node)
        no_indentation_node.add_node(another_one_indentation_node)

        assert one_indentation_node == no_indentation_node.children[0]
        assert two_indentation_node == no_indentation_node.children[0].children[0]
        assert another_one_indentation_node == no_indentation_node.children[1]

    def test_adds_multiple_nodes_to_one(self):
        start = self._read_node("%div")
        one = self._read_node("  %div")
        two = self._read_node("  %div")
        three = self._read_node("  %div")

        start.add_node(one)
        start.add_node(two)
        start.add_node(three)

        assert len(start.children) == 3

    @staticmethod
    def _read_node(haml):
        return read_node(Stream(haml), None, Compiler())



import codecs
import os
import time
import unittest
from os import listdir, path

import regex

from hamlpy.compiler import Compiler

TEMPLATE_DIRECTORY = "/templates/"
TEMPLATE_EXTENSION = ".hamlpy"


class TemplateCheck(object):
    compiler = Compiler(options={"format": "xhtml", "escape_attrs": True})

    def __init__(self, name, haml, html):
        self.name = name
        self.haml = haml
        self.expected_html = html.replace("\r", "")  # ignore line ending differences
        self.actual_html = None

    @classmethod
    def load_all(cls):
        directory = os.path.dirname(__file__) + TEMPLATE_DIRECTORY
        tests = []

        # load all test files
        for f in listdir(directory):
            haml_path = path.join(directory, f)
            if haml_path.endswith(TEMPLATE_EXTENSION):
                html_path = path.splitext(haml_path)[0] + ".html"

                haml = codecs.open(haml_path, encoding="utf-8").read()
                html = open(html_path, "r").read()

                tests.append(TemplateCheck(haml_path, haml, html))

        return tests

    def run(self):
        parsed = self.compiler.process(self.haml)

        # ignore line ending differences and blank lines
        self.actual_html = parsed.replace("\r", "")
        self.actual_html = regex.sub("\n[ \t]+(?=\n)", "\n", self.actual_html)

    def passed(self):
        return self.actual_html == self.expected_html


class TemplateCompareTest(unittest.TestCase):
    def test_templates(self):
        tests = TemplateCheck.load_all()

        for test in tests:
            print("Template test: " + test.name)
            test.run()

            if not test.passed():
                print("\nHTML (actual): ")
                print("\n".join(["%d. %s" % (i + 1, line) for i, line in enumerate(test.actual_html.split("\n"))]))
                self._print_diff(test.actual_html, test.expected_html)
                self.fail()

    @staticmethod
    def _print_diff(s1, s2):
        if len(s1) > len(s2):
            shorter = s2
        else:
            shorter = s1

        line = 1
        col = 1

        for i, _ in enumerate(shorter):
            if len(shorter) <= i + 1:
                print("Ran out of characters to compare!")
                print("Actual len=%d" % len(s1))
                print("Expected len=%d" % len(s2))
                break
            if s1[i] != s2[i]:
                print("Difference begins at line", line, "column", col)
                actual_line = s1.splitlines()[line - 1]
                expected_line = s2.splitlines()[line - 1]
                print("HTML (actual, len=%2d)   : %s" % (len(actual_line), actual_line))
                print("HTML (expected, len=%2d) : %s" % (len(expected_line), expected_line))
                print("Character code (actual)  : %d (%s)" % (ord(s1[i]), s1[i]))
                print("Character code (expected): %d (%s)" % (ord(s2[i]), s2[i]))
                break

            if shorter[i] == "\n":
                line += 1
                col = 1
            else:
                col += 1
        else:
            print("No Difference Found")


def performance_test(num_runs):
    """
    Performance test which evaluates all the testing templates a given number of times
    """
    print("Loading all templates...")

    template_tests = TemplateCheck.load_all()

    print("Running templates tests...")

    times = []

    for r in range(num_runs):
        start = time.time()

        for test in template_tests:
            test.run()

        times.append(time.time() - start)

    print(
        "Ran template tests %d times in %.2f seconds (average = %.3f secs)"
        % (num_runs, sum(times), sum(times) / float(num_runs))
    )


if __name__ == "__main__":
    performance_test(500)



import os
import shutil
import sys
import time
import unittest
from unittest.mock import patch

from hamlpy.hamlpy_watcher import watch_folder

WORKING_DIR = ".watcher_test"
INPUT_DIR = WORKING_DIR + os.sep + "input"
OUTPUT_DIR = WORKING_DIR + os.sep + "output"


class ScriptExit(Exception):
    def __init__(self, exit_code):
        self.exit_code = exit_code


class WatcherTest(unittest.TestCase):
    def test_watch_folder(self):
        # remove working directory if it exists and re-create it
        if os.path.exists(WORKING_DIR):
            shutil.rmtree(WORKING_DIR)

        os.makedirs(INPUT_DIR)

        # create some haml files for testing
        self._write_file(INPUT_DIR + os.sep + "test.haml", "%span{'class': 'test'}\n- shoutout\n")
        self._write_file(INPUT_DIR + os.sep + "error.haml", "%div{")
        self._write_file(INPUT_DIR + os.sep + ".#test.haml", "%hr")  # will be ignored

        # run as once off pass - should return 1 for number of failed conversions
        self._run_script(
            [
                "hamlpy_watcher.py",
                INPUT_DIR,
                OUTPUT_DIR,
                "--once",
                "--input-extension=haml",
                "--verbose",
                "--tag=shoutout:endshoutout",
                "--django-inline",
                "--jinja",
            ],
            1,
        )

        # check file without errors was converted
        self.assertFileContents(
            OUTPUT_DIR + os.sep + "test.html", "<span class='test'></span>\n{% shoutout %}\n{% endshoutout %}\n"
        )

        # check .#test.haml was ignored
        assert not os.path.exists(OUTPUT_DIR + os.sep + ".#test.html")

        # run without output directory which should make it default to re-using the input directory
        self._run_script(["hamlpy_watcher.py", INPUT_DIR, "--once", "--tag=shoutout:endshoutout"], 1)

        self.assertFileContents(
            INPUT_DIR + os.sep + "test.html", "<span class='test'></span>\n{% shoutout %}\n{% endshoutout %}\n"
        )

        # fix file with error
        self._write_file(INPUT_DIR + os.sep + "error.haml", "%div{}")

        # check exit code is zero
        self._run_script(["hamlpy_watcher.py", INPUT_DIR, OUTPUT_DIR, "--once"], 0)

        # run in watch mode with 1 second refresh
        self._run_script(
            ["hamlpy_watcher.py", INPUT_DIR, "--refresh=1", "--input-extension=haml", "--tag=shoutout:endshoutout"], 1
        )

    def assertFileContents(self, path, contents):
        with open(path, "r") as f:
            self.assertEqual(f.read(), contents)

    def _write_file(self, path, text):
        with open(path, "w") as f:
            f.write(text)

    def _run_script(self, script_args, expected_exit_code):
        def raise_exception_with_code(code):
            raise ScriptExit(code)

        # patch sys.exit so it throws an exception so we can return execution to this test
        # patch sys.argv to pass our arguments to the script
        # patch time.sleep to be interrupted
        with patch.object(sys, "exit", side_effect=raise_exception_with_code), patch.object(
            sys, "argv", script_args
        ), patch.object(time, "sleep", side_effect=KeyboardInterrupt), self.assertRaises(
            ScriptExit
        ) as raises:  # noqa
            watch_folder()

            assert raises.exception.exit_code == expected_exit_code



import os

import jinja2.ext

from hamlpy import HAML_EXTENSIONS
from hamlpy.compiler import Compiler
from hamlpy.parser.core import ParseException


class HamlPyExtension(jinja2.ext.Extension):
    def preprocess(self, source, name, filename=None):
        extension = os.path.splitext(name)[1][1:]

        if extension in HAML_EXTENSIONS:
            compiler = Compiler()
            try:
                return compiler.process(source)
            except ParseException as e:
                raise jinja2.TemplateSyntaxError(str(e), 1, name=name, filename=filename)
        else:
            return source



import django.views.generic

from hamlpy import HAML_EXTENSIONS

pouet = [
    "ArchiveIndexView",
    "YearArchiveView",
    "MonthArchiveView",
    "WeekArchiveView",
    "DayArchiveView",
    "TodayArchiveView",
    "DateDetailView",
    "DetailView",
    "CreateView",
    "UpdateView",
    "DeleteView",
    "ListView",
]

NON_HAML_EXTENSIONS = ("html", "htm", "xml")


class HamlExtensionTemplateView(object):
    def get_template_names(self):
        names = super(HamlExtensionTemplateView, self).get_template_names()

        haml_names = []

        for name in names:
            for ext in NON_HAML_EXTENSIONS:
                if name.endswith("." + ext):
                    base_name = name[: -len(ext)]

                    for haml_ext in HAML_EXTENSIONS:
                        haml_names.append(base_name + haml_ext)

        return haml_names + names


for view in pouet:
    locals()[view] = type(str(view), (HamlExtensionTemplateView, getattr(django.views.generic, view)), {})



import ast

STRING_LITERALS = ('"', "'")
WHITESPACE_CHARS = (" ", "\t")
WHITESPACE_AND_NEWLINE_CHARS = (" ", "\t", "\r", "\n")


class ParseException(Exception):
    def __init__(self, message, stream=None):
        if stream:
            context = stream.text[max(stream.ptr - 31, 0) : stream.ptr + 1]
            message = '%s @ "%s" <-' % (message, context)

        super(ParseException, self).__init__(message)


class Stream(object):
    def __init__(self, text):
        self.text = text
        self.length = len(self.text)
        self.ptr = 0

    def expect_input(self):
        """
        We expect more input, raise exception if there isn't any
        """
        if self.ptr >= self.length:
            raise ParseException("Unexpected end of input.", self)

    def raise_unexpected(self):
        """
        Raises exception that current character is unexpected
        """
        raise ParseException('Unexpected "%s".' % self.text[self.ptr], self)

    def __repr__(self):  # pragma: no cover
        return '"%s" >> "%s"' % (
            self.text[: self.ptr].replace("\n", "\\n"),
            self.text[self.ptr :].replace("\n", "\\n"),
        )


class TreeNode(object):
    """
    Generic parent/child tree class
    """

    def __init__(self):
        self.parent = None
        self.children = []

    def left_sibling(self):
        siblings = self.parent.children
        index = siblings.index(self)
        return siblings[index - 1] if index > 0 else None

    def right_sibling(self):
        siblings = self.parent.children
        index = siblings.index(self)
        return siblings[index + 1] if index < len(siblings) - 1 else None

    def add_child(self, child):
        child.parent = self
        self.children.append(child)


def read_whitespace(stream, include_newlines=False):
    """
    Reads whitespace characters, returning the whitespace characters
    """
    whitespace = WHITESPACE_AND_NEWLINE_CHARS if include_newlines else WHITESPACE_CHARS

    start = stream.ptr

    while stream.ptr < stream.length and stream.text[stream.ptr] in whitespace:
        stream.ptr += 1

    return stream.text[start : stream.ptr]


def peek_indentation(stream):
    """
    Counts but doesn't actually read indentation level on new line, returning the count or None if line is blank
    """
    indentation = 0
    while True:
        ch = stream.text[stream.ptr + indentation]
        if ch == "\n":
            return None

        if not ch.isspace():
            return indentation

        indentation += 1


def read_quoted_string(stream):
    """
    Reads a single or double quoted string, returning the value without the quotes
    """
    terminator = stream.text[stream.ptr]

    assert terminator in STRING_LITERALS

    start = stream.ptr
    stream.ptr += 1  # consume opening quote

    while True:
        if stream.ptr >= stream.length:
            raise ParseException("Unterminated string (expected %s)." % terminator, stream)

        if stream.text[stream.ptr] == terminator and stream.text[stream.ptr - 1] != "\\":
            break

        stream.ptr += 1

    stream.ptr += 1  # consume closing quote

    # evaluate as a Python string (evaluates escape sequences)
    return ast.literal_eval(stream.text[start : stream.ptr])


def read_line(stream):
    """
    Reads a line
    """
    start = stream.ptr

    if stream.ptr >= stream.length:
        return None

    while stream.ptr < stream.length and stream.text[stream.ptr] != "\n":
        stream.ptr += 1

    line = stream.text[start : stream.ptr]

    if stream.ptr < stream.length and stream.text[stream.ptr] == "\n":
        stream.ptr += 1

    return line


def read_number(stream):
    """
    Reads a decimal number, returning value as string
    """
    start = stream.ptr

    while True:
        if not stream.text[stream.ptr].isdigit() and stream.text[stream.ptr] != ".":
            break

        stream.ptr += 1

    return stream.text[start : stream.ptr]


def read_symbol(stream, symbols):
    """
    Reads one of the given symbols, returning its value
    """
    for symbol in symbols:
        if stream.text[stream.ptr : stream.ptr + len(symbol)] == symbol:
            stream.ptr += len(symbol)
            return symbol

    raise ParseException("Expected %s." % " or ".join(['"%s"' % s for s in symbols]), stream)


def read_word(stream, include_chars=()):
    """
    Reads a sequence of word characters
    """
    stream.expect_input()

    start = stream.ptr

    while stream.ptr < stream.length:
        ch = stream.text[stream.ptr]
        if not (ch.isalnum() or ch == "_" or ch in include_chars):
            break
        stream.ptr += 1

    # if we immediately hit a non-word character, raise it as unexpected
    if start == stream.ptr:
        stream.raise_unexpected()

    return stream.text[start : stream.ptr]



import os

from django.conf import settings
from django.template import TemplateDoesNotExist
from django.template.loaders import app_directories, filesystem

from hamlpy import HAML_EXTENSIONS
from hamlpy.compiler import Compiler
from hamlpy.template.utils import get_django_template_loaders

# Get options from Django settings
options = {}

if hasattr(settings, "HAMLPY_ATTR_WRAPPER"):
    options.update(attr_wrapper=settings.HAMLPY_ATTR_WRAPPER)

if hasattr(settings, "HAMLPY_DJANGO_INLINE_STYLE"):
    options.update(django_inline_style=settings.HAMLPY_DJANGO_INLINE_STYLE)


def get_haml_loader(loader):
    class Loader(loader.Loader):
        def get_contents(self, origin):
            # Django>=1.9
            contents = super(Loader, self).get_contents(origin)
            name, _extension = os.path.splitext(origin.template_name)
            # os.path.splitext always returns a period at the start of extension
            extension = _extension.lstrip(".")

            if extension in HAML_EXTENSIONS:
                compiler = Compiler(options=options)
                return compiler.process(contents)

            return contents

        def load_template_source(self, template_name, *args, **kwargs):
            # Django<1.9
            name, _extension = os.path.splitext(template_name)
            # os.path.splitext always returns a period at the start of extension
            extension = _extension.lstrip(".")

            if extension in HAML_EXTENSIONS:
                try:
                    haml_source, template_path = super(Loader, self).load_template_source(
                        self._generate_template_name(name, extension), *args, **kwargs
                    )
                except TemplateDoesNotExist:  # pragma: no cover
                    pass
                else:
                    compiler = Compiler(options=options)
                    html = compiler.process(haml_source)

                    return html, template_path

            raise TemplateDoesNotExist(template_name)

        load_template_source.is_usable = True

        @staticmethod
        def _generate_template_name(name, extension="hamlpy"):
            return "%s.%s" % (name, extension)

    return Loader


haml_loaders = dict((name, get_haml_loader(loader)) for (name, loader) in get_django_template_loaders())

HamlPyFilesystemLoader = get_haml_loader(filesystem)
HamlPyAppDirectoriesLoader = get_haml_loader(app_directories)



import importlib
from unittest import mock

import pytest

from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.test import SimpleTestCase
from django.test.utils import override_settings

from hamlpy.compiler import Compiler
from hamlpy.template import loaders


class LoaderTest(SimpleTestCase):
    def setUp(self):
        super(LoaderTest, self).setUp()

        importlib.reload(loaders)

    @mock.patch("hamlpy.template.loaders.Compiler", wraps=Compiler)
    def test_compiler_default_settings(self, mock_compiler_class):
        render_to_string("simple.hamlpy")

        mock_compiler_class.assert_called_once_with(options={})
        mock_compiler_class.reset_mock()

    @override_settings(HAMLPY_ATTR_WRAPPER='"', HAMLPY_DJANGO_INLINE_STYLE=False)
    def test_compiler_settings(self):
        importlib.reload(loaders)

        with mock.patch("hamlpy.template.loaders.Compiler", wraps=Compiler) as mock_compiler_class:
            rendered = render_to_string("simple.hamlpy")

            mock_compiler_class.assert_called_once_with(options={"attr_wrapper": '"', "django_inline_style": False})

            assert '"someClass"' in rendered

    def test_template_rendering(self):
        assert render_to_string("simple.hamlpy") == self._load_test_template("simple.html")

        context = {
            "section": {"title": "News", "subtitle": "Technology"},
            "story_list": [
                {
                    "headline": "Haml Helps",
                    "tease": "Many HAML users...",
                    "get_absolute_url": lambda: "http://example.com/stories/1/",
                }
            ],
        }

        rendered = render_to_string("djangoCombo.hamlpy", context)

        assert "<h2>Technology</h2>" in rendered
        assert "HAML HELPS" in rendered
        assert "<a href='http://example.com/stories/1/'>" in rendered
        assert "<p>Many HAML users...</p>"

    def test_should_ignore_non_haml_templates(self):
        assert render_to_string("simple.html") == self._load_test_template("simple.html")

    def test_should_raise_exception_when_template_doesnt_exist(self):
        with pytest.raises(TemplateDoesNotExist):
            render_to_string("simple.xyz")

    def _load_test_template(self, name):
        return open("hamlpy/test/templates/" + name, "r").read()



from importlib import machinery
from os import listdir
from os.path import dirname, splitext

from django.template import loaders

MODULE_EXTENSIONS = tuple(machinery.all_suffixes())


def get_django_template_loaders():
    return [
        (loader.__name__.rsplit(".", 1)[1], loader) for loader in get_submodules(loaders) if hasattr(loader, "Loader")
    ]


def get_submodules(package):
    submodules = ("%s.%s" % (package.__name__, module) for module in package_contents(package))
    return [__import__(module, {}, {}, [module.rsplit(".", 1)[-1]]) for module in submodules]


def package_contents(package):
    package_path = dirname(loaders.__file__)
    contents = set([splitext(module)[0] for module in listdir(package_path) if module.endswith(MODULE_EXTENSIONS)])
    return contents



from collections import OrderedDict

import regex

from .core import (
    STRING_LITERALS,
    WHITESPACE_CHARS,
    ParseException,
    peek_indentation,
    read_line,
    read_number,
    read_quoted_string,
    read_symbol,
    read_whitespace,
    read_word,
)
from .utils import html_escape

LEADING_SPACES_REGEX = regex.compile(r"^\s+", regex.V1 | regex.MULTILINE)

# non-word characters that we allow in attribute keys (in HTML style attribute dicts)
ATTRIBUTE_KEY_EXTRA_CHARS = {":", "-", "$", "?", "[", "]"}

ATTRIBUTE_VALUE_KEYWORDS = {"none": None, "true": True, "false": False}


def read_attribute_value(stream, compiler):
    """
    Reads an attribute's value which may be a string, a number or None
    """
    ch = stream.text[stream.ptr]

    if ch in STRING_LITERALS:
        value = read_quoted_string(stream)

        if compiler.options.escape_attrs:
            # TODO handle escape_attrs=once
            value = html_escape(value)

    elif ch.isdigit():
        value = read_number(stream)
    else:
        raw_value = read_word(stream)

        if raw_value.lower() in ATTRIBUTE_VALUE_KEYWORDS:
            value = ATTRIBUTE_VALUE_KEYWORDS[raw_value.lower()]
        else:
            value = "{{ %s }}" % raw_value

    return value


def read_attribute_value_list(stream, compiler):
    """
    Reads an attribute value which is a list of other values
    """
    open_literal = stream.text[stream.ptr]

    assert open_literal in ("(", "[")

    read_tuple = open_literal == "("
    close_literal = ")" if read_tuple else "]"

    data = []

    stream.ptr += 1  # consume opening symbol

    while True:
        read_whitespace(stream)

        if stream.text[stream.ptr] == close_literal:
            break

        data.append(read_attribute_value(stream, compiler))

        read_whitespace(stream)

        if stream.text[stream.ptr] != close_literal:
            read_symbol(stream, (",",))

    stream.ptr += 1  # consume closing symbol

    return data


def read_attribute_value_haml(stream, compiler):
    """
    Reads an attribute value which is a block of indented Haml
    """
    indentation = peek_indentation(stream)
    haml_lines = []

    # read lines below with higher indentation as this filter's content
    while stream.ptr < stream.length:
        line_indentation = peek_indentation(stream)

        if line_indentation is not None and line_indentation < indentation:
            break

        line = read_line(stream)

        # don't preserve whitespace on empty lines
        if line.isspace():
            line = ""

        haml_lines.append(line)

    stream.ptr -= 1  # un-consume final newline which will act as separator between this and next entry

    haml = "\n".join(haml_lines)
    html = compiler.process(haml)

    # un-format into single line
    return LEADING_SPACES_REGEX.sub(" ", html).replace("\n", "").strip()


def read_ruby_attribute(stream, compiler):
    """
    Reads a Ruby style attribute, e.g. :foo => "bar" or foo: "bar"
    """
    old_style = stream.text[stream.ptr] == ":"

    if old_style:
        stream.ptr += 1

        key = read_word(stream, include_chars=("-",))
    else:
        # new style Ruby / Python style allows attribute to be quoted string
        if stream.text[stream.ptr] in STRING_LITERALS:
            key = read_quoted_string(stream)

            if not key:
                raise ParseException("Attribute name can't be an empty string.", stream)
        else:
            key = read_word(stream, include_chars=("-",))

    read_whitespace(stream)

    if stream.text[stream.ptr] in ("=", ":"):
        if old_style:
            read_symbol(stream, ("=>",))
        else:
            read_symbol(stream, (":",))

        read_whitespace(stream, include_newlines=False)

        stream.expect_input()

        if stream.text[stream.ptr] == "\n":
            stream.ptr += 1

            value = read_attribute_value_haml(stream, compiler)
        elif stream.text[stream.ptr] in ("(", "["):
            value = read_attribute_value_list(stream, compiler)
        else:
            value = read_attribute_value(stream, compiler)
    else:
        value = True

    return key, value


def read_html_attribute(stream, compiler):
    """
    Reads an HTML style attribute, e.g. foo="bar"
    """
    key = read_word(stream, include_chars=ATTRIBUTE_KEY_EXTRA_CHARS)

    # can't have attributes without whitespace separating them
    ch = stream.text[stream.ptr]
    if ch not in WHITESPACE_CHARS and ch not in ("=", ")"):
        stream.raise_unexpected()

    read_whitespace(stream)

    if stream.text[stream.ptr] == "=":
        read_symbol(stream, "=")

        read_whitespace(stream, include_newlines=False)

        stream.expect_input()

        if stream.text[stream.ptr] == "\n":
            stream.ptr += 1

            value = read_attribute_value_haml(stream, compiler)
        elif stream.text[stream.ptr] == "[":
            value = read_attribute_value_list(stream, compiler)
        else:
            value = read_attribute_value(stream, compiler)
    else:
        value = True

    return key, value


def read_attribute_dict(stream, compiler):
    """
    Reads an attribute dictionary which may use one of 3 syntaxes:
     1. {:foo => "bar", :a => 3}  (old Ruby)
     2. {foo: "bar", a: 3}  (new Ruby / Python)
     3. (foo="bar" a=3)  (HTML)
    """
    data = OrderedDict()

    opener = stream.text[stream.ptr]

    assert opener in ("{", "(")

    if opener == "(":
        html_style = True
        terminator = ")"
    else:
        html_style = False
        terminator = "}"

    stream.ptr += 1

    def record_value(key, value):
        if key in data:
            raise ParseException('Duplicate attribute: "%s".' % key, stream)
        data[key] = value

    while True:
        read_whitespace(stream, include_newlines=True)

        if stream.ptr >= stream.length:
            raise ParseException("Unterminated attribute dictionary", stream)

        if stream.text[stream.ptr] == terminator:
            stream.ptr += 1
            break

        # (foo = "bar" a=3)
        if html_style:
            record_value(*read_html_attribute(stream, compiler))

            read_whitespace(stream)

            if stream.text[stream.ptr] == ",":
                raise ParseException('Unexpected ",".', stream)

        # {:foo => "bar", :a=>3} or {foo: "bar", a: 3}
        else:
            record_value(*read_ruby_attribute(stream, compiler))

            read_whitespace(stream)

            if stream.text[stream.ptr] not in (terminator, "\n"):
                read_symbol(stream, (",",))

    return data
